//=================================
// include guard
#pragma once

//=================================
// included dependencies from EPSO small
#include "fitnessfunction.h"

//=================================
// included dependencies from ITL networks
#include "..\ITL_networks\General_Functions.h"
#include "..\ITL_networks\Normalization_functions.h"
#include "..\enumerations.h"
#include "..\Hybrid_functions\hyb_functions.h"
//=================================
// included dependencies
#include <armadillo>

//=================================
// included namespaces
using namespace std;
using namespace arma;

//=================================
// the HydroWindHWS class
class HydroWindHWS : public fitnessfunction
{
public:
	HydroWindHWS():fitnessfunction(){ }; // constructor

	double calcFitness(
		const fvec &variaveisProblema,   
		const vec &typeperiodEntry, 
		const vec &windgenEntry, 
		vec xprefinalEntry, 
		vec xfinalEntry,
		const vec &capmin, 
		const vec &capmax, 
		const vec &inflow, 
		const vec &numres, 
		const vec &qmin, 
		const vec &qmax, 
		vector <double> Y, 
		const mat &W1,
		const mat &W2,
		const rowvec &B1,
		const rowvec &B2,  const rowvec &Parameter1, const rowvec &Parameter2,
		const float &actFcnFlatness, 
		const ActivationFunctions &NeuronActFunct2,
		const NormalizationMethod &NormMethod,
		ProblemSelection ProblemSelect)
	{
		float numR = variaveisProblema(12);
		unsigned int DimHidden = variaveisProblema (14);
		unsigned int DimIn=numR*variaveisProblema(13);

		//=================================
		//expansion and desnormalization
		mat PatternOriginal (1,DimHidden);
		for (unsigned int k=0; k<DimHidden; k++)
		{
			PatternOriginal (0,k) = Y[k];
		}
		unsigned short numPatternsDesteCaso=1;

		mat PatternExpandido (1,DimIn);
		General_functions::transitionLayer (PatternExpandido, numPatternsDesteCaso, DimHidden, DimIn, PatternOriginal, W2, B2, actFcnFlatness, NeuronActFunct2);

		mat PatternDESnormalized (1,DimIn);

		PatternDESnormalized = Normalization::DES_Normalization_function (Parameter1, Parameter2, PatternExpandido, numPatternsDesteCaso, DimIn, NormMethod);

		vector <double> Pos (DimIn); 
		for (unsigned int k=0; k<DimIn; k++)
		{
			Pos[k]= PatternDESnormalized(0,k);
		}

				
		vec xfinal = xfinalEntry;
		vec xprefinal = xprefinalEntry;

		vec xanterior(DimIn);
		vec z (DimIn);
		
		
		//=================================
		//limites dinamicos da particula
		for (unsigned int i=0; i<DimIn; i++)
		{
			xanterior[i]=xfinal[i];

			//gama
			double gama = HYB_func::Gama_func(i, numres, xfinal, capmin, ProblemSelect);

			//construo de sumomega 
			double sumomega = HYB_func::sumOmega_func(i, numres, z, Pos, ProblemSelect);

			//=================================
			//minPos e maxPos
			double maxPos;
			double minPos; 
						
			HYB_func::MinMaxPos	(
					maxPos,
					minPos,		
					i, 
					xanterior,
					Pos,
					capmin,
					capmax,
					inflow, 
					numres, 
					qmin, 
					qmax, 
					sumomega,
					gama,
					ProblemSelect);
			
			if (Pos[i]<minPos)
			{
				Pos[i]=minPos;
			}
			
			if (Pos[i]>maxPos)
			{
				Pos[i]=maxPos;
			}			

			//xprefinal
			xprefinal[i+numR]=xanterior[i] + inflow[i] + sumomega - Pos[i];
		

			//obriga a respeitar a capmin
			if (xprefinal[i+numR]<capmin[i])
			{
				Pos[i]-=capmin[i]-xprefinal[i+numR];
				
				//Pos[i] = Pos[i] - capmin[i] + xprefinal[i+num];
				xprefinal[i+numR]=xanterior[i]+inflow[i]+sumomega - Pos[i];
			}

			//vetor z
			z[i]=max(0.0,(xprefinal[i+numR]-capmax[i])); 

			//xfinal
			xfinal[i+numR]=xprefinal[i+numR]-z[i];
		}

		

		double fit = HYB_func::ReturnFit (
				Pos,
				variaveisProblema,
				typeperiodEntry, 
				windgenEntry, 
				xanterior,
				xfinal);
			
	
		return (fit);
	}
};

