//=================================
// include guard
#pragma once

//=================================
// forward declared dependencies
class fitnessfunction;

//=================================
// included dependencies
//#include <vector>
//#include <iostream>
//#include <cassert>
//#include <ctime>
#include <armadillo>

#include <boost/random.hpp>
#include <boost/random/uniform_real.hpp>
#include <boost/random/lognormal_distribution.hpp>

#include "fitnessfunction.h"

#include "..\enumerations.h"
#include "..\ITL_networks\General_Functions.h"
#include "..\ITL_networks\Normalization_functions.h"
#include "..\enumerations.h"

//=================================
// namespaces
using namespace std;
using namespace arma;

//=================================
// the particle class
class particle
{
public:
	//=================================
	~particle(void); // destructor

	//=================================
	particle(void); // constructor, default
	
	//=================================
	particle(
		const fvec &variaveisProblema,   
		const vec &typeperiodEntry, 
		const vec &windgenEntry, 
		vec xprefinalEntry, 
		vec xfinalEntry,
		const vec &capmin, 
		const vec &capmax, 
		const vec &inflow, 
		const vec &numres, 
		const vec &qmin, 
		const vec &qmax, 
		vector<double> tempPos,
		vector<double> tempBeforePos,
		vector<double> tempBestPos,
		const mat &W1, const mat &W2,
		const rowvec &B1, const rowvec &B2,  const rowvec &Parameter1, const rowvec &Parameter2,
		const float &actFcnFlatness,
		const ActivationFunctions &NeuronActFunct1,
		const ActivationFunctions &NeuronActFunct2,
		const NormalizationMethod &NormMethod,
		double inertia,
		double memory,
		double cooperation,
		double perturbation,
		boost::mt19937& rng,
		fitnessfunction *pFf, 
		const double &CpP,
		ProblemSelection ProblemSelect,
		const Initi_Vel &Vel_init);
	
	//=================================	
	void Print(); // dump the information about the particle, can be done by overloading the <<operator, for now this will do
	
	//=================================
	double GetFitness(); // gets the current fitness value but does not recalculate fitness each time
	
	//=================================
	void UpdateFitness(
		const fvec &variaveisProblema,   
		const vec &typeperiodEntry, 
		const vec &windgenEntry, 
		vec xprefinalEntry, 
		vec xfinalEntry,
		const vec &capmin, 
		const vec &capmax, 
		const vec &inflow, 
		const vec &numres, 
		const vec &qmin, 
		const vec &qmax, 
		const mat &W1, const mat &W2,
		const rowvec &B1, const rowvec &B2,  const rowvec &Parameter1, const rowvec &Parameter2,
		const float &actFcnFlatness, 
		const ActivationFunctions &NeuronActFunct2,
		const NormalizationMethod &NormMethod,
		ProblemSelection ProblemSelect); // update the fitness

	//=================================
	vector <double > GetBestPos(); //get the particle's own previous best position (pBest)
	
	//=================================
	void SetBestPos(vector<double > newGlobalBestPos); //set global best position gBest for this particle
	
	//=================================
	vector <double > GetPos(); //get particle's current position (location, vector X)
	
	//=================================
	vector <double > GetVel(); //get particle's current velocity
	
	//=================================
	void Mutate(
		boost::mt19937& rng, 
		const double &TauP); // mutate the internal parameters (EPSO)
	
	//=================================
	void Movement(
		const fvec &variaveisProblema,   
		const vec &typeperiodEntry, 
		const vec &windgenEntry, 
		vec xprefinalEntry, 
		vec xfinalEntry,
		const vec &capmin, 
		const vec &capmax, 
		const vec &inflow, 
		const vec &numres, 
		const vec &qmin, 
		const vec &qmax, 
		boost::mt19937& rng, 
		const mat &W1, const mat &W2,
		const rowvec &B1, const rowvec &B2,  const rowvec &Parameter1, const rowvec &Parameter2,
		const float &actFcnFlatness,
		const ActivationFunctions &NeuronActFunct2,
		const NormalizationMethod &NormMethod,
		const int &DimHidden,
		const double &CpP,
		ProblemSelection ProblemSelect);
	
		

	//=================================
	double GetSInertia();

	//=================================
	double GetSMemory();

	//=================================
	double GetSCooperation();

	//=================================
	double GetSPerturbation();
	
protected:
		
	double Fitness; // stores current fitness
	vector<double > Pos; // position
	vector<double > Vel; // velocity
	vector<double > myBestPos; // particle previous best, pBest
	vector<double > globalBestPos; //global best, gBest
	double myBestFitness; //particle's best fitness in myBestPos (pBest)
	
	
	double inertiaParteB;
	double memoryParteB;
	double cooperationParteB;
	double perturbationParteB;

	// EPSO weights

	fitnessfunction *ff; // pointer to fitness function

	vector <double> maxPos;
	vector <double> minPos;
	vector <double> maxVel;
	vector <double> minVel; // limits for position and velocity

	
	
	vector<double> capacidadeMin, normCapacidadeMin;
};
