
#include "swarm.h"


swarm::~swarm(void){}

//construtor do swarm do espao reduzido
swarm::swarm(
	const fvec &variaveisProblema,   
	const vec &typeperiodEntry, 
	const vec &windgenEntry, 
	vec xprefinalEntry, 
	vec xfinalEntry,
	const vec &capmin, 
	const vec &capmax, 
	const vec &inflow, 
	const vec &numres, 
	const vec &qmin, 
	const vec &qmax, 
	list <HWparticle> &lastPrimeiroSwarm,
	list <HWparticle> &beforeLastPrimeiroSwarm,
	const int &numParticles, 
	const mat &W1, const mat &W2,
	const rowvec &B1, const rowvec &B2, const rowvec &Parameter1, const rowvec &Parameter2,
	float actFcnFlatness,
	const ActivationFunctions &NeuronActFunct1,
	ActivationFunctions NeuronActFunct2,
	NormalizationMethod NormMethod,
	bool newMinimize, 
	fitnessfunction &newFf, 
	boost::mt19937& rng, 
	const double &CpP, 
	const double &TauP,
	ProblemSelection ProblemSelect,
	const Initi_Vel &Vel_init) : ff(newFf)
{
	this->numEvalsParteB = 0;
	this->minimize = newMinimize;
	this->ff = newFf;
		
	float numR = variaveisProblema(12);
	unsigned int DimIn=numR*variaveisProblema(13);
	

	ff.SetMinimize(minimize);  
	
	list<HWparticle>::iterator pParticle;
	list<HWparticle>::iterator beforeParticle;
	pParticle=lastPrimeiroSwarm.begin();
	beforeParticle=beforeLastPrimeiroSwarm.begin();

	while(pParticle!=lastPrimeiroSwarm.end())
	{
		vector<double> tempPos, tempBeforePos, tempBestPos;
		for (unsigned int i=0;i<DimIn;i++)
		{
			tempPos.push_back(pParticle->HWGetPos()[i]);
			tempBestPos.push_back(pParticle->HWGetBestPos()[i]);

			tempBeforePos.push_back(beforeParticle->HWGetPos()[i]);
		}

		//Fitness=pParticle->HWGetFitness();
		inertiaParteB=(pParticle->GetInertia());
		memoryParteB=(pParticle->GetMemory());
		cooperationParteB=(pParticle->GetCooperation());
		perturbationParteB=(pParticle->GetPerturbation());

#if 0 //#if 1 reinicializa os pesos (gaussian mutation)
		boost::normal_distribution<double> norm_dist(0.0, 1.0);
		boost::variate_generator<boost::mt19937&, boost::normal_distribution<double> >  rng_sampler(rng, norm_dist);
						
		Inertia=rng_sampler();
		Memory=rng_sampler();
		Cooperation=rng_sampler();
		Perturbation=rng_sampler();
#endif
				
		EnxameParteB.push_back(particle(
			variaveisProblema,  
			typeperiodEntry, 
			windgenEntry, 
			xprefinalEntry, 
			xfinalEntry,
			capmin, 
			capmax, 
			inflow, 
			numres, 
			qmin, 
			qmax, 
			tempPos, 
			tempBeforePos, 
			tempBestPos,
			W1, W2,	B1,	B2, Parameter1, Parameter2,
			actFcnFlatness,
			NeuronActFunct1,
			NeuronActFunct2,
			NormMethod,
			inertiaParteB, 
			memoryParteB, 
			cooperationParteB, 
			perturbationParteB,
			rng, 
			&ff, 
			CpP,
			ProblemSelect, 
			Vel_init));
				
		pParticle++;
		beforeParticle++;
	}
	

	EnxameParteB.sort(ff.particleCompare); // may be stupid: sort list of particles according to their fitness

	this->gBestFitness = EnxameParteB.begin()->GetFitness();
	this->gBestPos = EnxameParteB.begin()->GetPos();  // get the global best for the beginning
}

unsigned int swarm::GetNumEvalsParteB()
{
	return this->numEvalsParteB;
}

double swarm::GetBestFitness()
{
	return gBestFitness;
}

vector<double> swarm::GetBestPos()
{
	return gBestPos;
}

void swarm::NextIterationSegundo(
	const fvec &variaveisProblema,   
	const vec &typeperiodEntry, 
	const vec &windgenEntry,
	vec xprefinalEntry, 
	vec xfinalEntry,
	const vec &capmin, 
	const vec &capmax, 
	const vec &inflow, 
	const vec &numres, 
	const vec &qmin, 
	const vec &qmax, 
	boost::mt19937& rng, 
	const mat &W1, const mat &W2,
	const rowvec &B1, const rowvec &B2, const rowvec &Parameter1, const rowvec &Parameter2,
	float actFcnFlatness, 
	const ActivationFunctions &NeuronActFunct2,
	const NormalizationMethod &NormMethod,
	unsigned int maxIterS, 
	const unsigned int &numberSegundo,
	const int &DimHidden,
	const double &CpP,
	const double &TauP,
	ProblemSelection ProblemSelect)
{
	particle tempParticle;

	list<particle>::iterator pParticle;
	pParticle=EnxameParteB.begin();
		
	if (ff.fitnessCompare(pParticle->GetFitness(), gBestFitness))
	{
		gBestPos = pParticle->GetPos(); // new gBest found - update.
		gBestFitness = pParticle->GetFitness(); 
	}
	
	//estes minimos e maximos so para definir E LIMITAR o domnio de procura
	
#if 0
	int innerDim;
	int j=0;
	delta=0.5;
	innerDim=pParticle->GetPos().size();
	maxPos.resize(innerDim); minPos.resize(innerDim); maxVel.resize(innerDim); minVel.resize(innerDim);
	
	j=0;

	for(j=0; j<innerDim; j++)
		{
			while(pParticle!=listSmallParticles.end())
			{
				if (pParticle->GetPos()[j]>maxPos[j]) 
				{maxPos[j]=delta+(pParticle->GetPos()[j]);}
					
				if (pParticle->GetPos()[j]<minPos[j])
				{minPos[j]=(pParticle->GetPos()[j])-delta;}

#if 0 //#if 1 calcula minVel and maxVel apartir da velocidade das particulas
				if (pParticle->GetVel()[j]>maxVel[j]) 
				{maxVel[j]=delta+(pParticle->GetVel()[j]);}
					
				if (pParticle->GetVel()[j]<minVel[j])
				{minVel[j]=delta+(pParticle->GetVel()[j]);}
#endif
					
				pParticle++;
			}
			pParticle=listSmallParticles.begin();
		}
#endif
#if 0
	for (j=0;j<innerDim;j++)
	{minPos[j]=-1;
	maxPos[j]=1;}
#endif
#if 0 //#if 1 calcula minVel e maxVel a partir das posies minimas e mximas dos vectores (limites do espao de busca)
	j=0;
	for (j=0; j<innerDim; j++)
	{
		minVel[j]=(-fabs(maxPos[j] - minPos[j]));
		maxVel[j]=(fabs(maxPos[j] - minPos[j]));
	}
#endif
	
	

	while(pParticle!=EnxameParteB.end())
	{
		pParticle->SetBestPos(gBestPos); // new gBest set
		tempParticle=*pParticle; // replicate particle, make a copy

		tempParticle.Mutate(rng, TauP); //  tau = 0.2, mutate internal parameters
		tempParticle.Movement(
			variaveisProblema,  
			typeperiodEntry, 
			windgenEntry, 
			xprefinalEntry, 
			xfinalEntry,
			capmin, 
			capmax, 
			inflow, 
			numres, 
			qmin, 
			qmax, 
			rng, W1, W2, B1, B2, Parameter1, Parameter2, 
			actFcnFlatness, NeuronActFunct2, NormMethod, DimHidden, CpP, ProblemSelect); // move the replicated particle
		
		(*pParticle).Movement(
			variaveisProblema,  
			typeperiodEntry, 
			windgenEntry, 
			xprefinalEntry, 
			xfinalEntry,
			capmin, 
			capmax, 
			inflow, 
			numres, 
			qmin, 
			qmax, 
			rng, W1, W2, B1, B2, Parameter1, Parameter2,
			actFcnFlatness, NeuronActFunct2, NormMethod, DimHidden, CpP, ProblemSelect); // move the original particle

		numEvalsParteB++;
		numEvalsParteB++;

		if ( ff.particleCompare(*pParticle, tempParticle) )
		{ // original particle chosen - nothing to do
		}
		else
		{ // mutated particle chosen
			(*pParticle)=tempParticle;
		}	
		pParticle++; // iterate through all the particles
	}
	EnxameParteB.sort(ff.particleCompare); // order the list of particles


	// iteration finished
	
	//==================================
	//copy of swarm at last iteration
	if (numberSegundo==maxIterS) //nota
	{
		pParticle=EnxameParteB.begin();
		while(pParticle!=EnxameParteB.end())
		{
			tempParticle=*pParticle;
			lastSegundoSwarm.push_back(tempParticle);
			pParticle++;
		}
	}

	//===================================
	//copy of swarm at t-1 iteration
	if (numberSegundo==maxIterS-1) //nota
	{
		pParticle=EnxameParteB.begin();
		while(pParticle!=EnxameParteB.end())
		{
			tempParticle=*pParticle;
			beforeLastSegundoSwarm.push_back(tempParticle);
			pParticle++;
		}
	}



}
