//=================================
// include guard
#pragma once

//=================================
// included dependencies
#include <iostream>
#include <fstream>
#include <iomanip>
#include <ctime>
#include <list>
#include <vector>

#include <boost/random.hpp>

#include "particle.h"
#include "fitnessfunction.h"
#include "../HW/particleHW.h"

#include "..\enumerations.h"

#include <armadillo>
//#include "..\ITL_networks\General_Functions.h"


//=================================
// namespaces
using namespace std;
using namespace arma;

//=================================
// the swarm class
class swarm
{
public:
	//=================================
	~swarm(void); // destructor
	
	//=================================
	swarm(
		const fvec &variaveisProblema,   
		const vec &typeperiodEntry, 
		const vec &windgenEntry, 
		vec xprefinalEntry, 
		vec xfinalEntry,
		const vec &capmin, 
		const vec &capmax, 
		const vec &inflow, 
		const vec &numres, 
		const vec &qmin, 
		const vec &qmax, 
		list <HWparticle> &lastPrimeiroSwarm,
		list <HWparticle> &beforeLastPrimeiroSwarm,
		const int &numParticles, 
		const mat &W1, const mat &W2,
		const rowvec &B1, const rowvec &B2, const rowvec &Parameter1, const rowvec &Parameter2,
		float actFcnFlatness,
		const ActivationFunctions &NeuronActFunct1,
		ActivationFunctions NeuronActFunct2,
		NormalizationMethod NormMethod,
		bool minimize,
		fitnessfunction &ff, 
		boost::mt19937 &rng, 
		const double &CpP, 
		const double &TauP,
		ProblemSelection ProblemSelect,
		const Initi_Vel &Vel_init); // constructor

	//=================================
	void NextIterationSegundo(
		const fvec &variaveisProblema,   
		const vec &typeperiodEntry, 
		const vec &windgenEntry, 
		vec xprefinalEntry, 
		vec xfinalEntry,
		const vec &capmin, 
		const vec &capmax, 
		const vec &inflow, 
		const vec &numres, 
		const vec &qmin, 
		const vec &qmax, 
		boost::mt19937& rng, 
		const mat &W1,
		const mat &W2,
		const rowvec &B1,
		const rowvec &B2, const rowvec &Parameter1, const rowvec &Parameter2,
		float actFcnFlatness, 
		const ActivationFunctions &NeuronActFunct2,
		const NormalizationMethod &NormMethod,
		unsigned int maxIterS, 
		const unsigned int &numberSegundo,
		const int &DimHidden,
		const double &CpP,
		const double &TauP,
		ProblemSelection ProblemSelect);
	
	//=================================
	unsigned int GetNumEvalsParteB();
	
	//=================================
	double GetBestFitness();

	//=================================
	vector <double> GetBestPos();

	//=================================
	//list holding the last swarm for the main program (last solution)
	list <particle> lastSegundoSwarm; 

	//=================================
	//list holding the swarm at t-1 (before last solution)
	list <particle> beforeLastSegundoSwarm; 
	
protected:

	double gBestFitness; // global best's fitness
	vector<double > gBestPos; // global best position
			
	list<particle> EnxameParteB;  // list holding the swarm

	unsigned int numEvalsParteB; // number of fitness function evaluations performed
	fitnessfunction &ff;
		
	bool minimize; // minimizing or maximizing
	
	double Fitness;

	double inertiaParteB;
	double memoryParteB;
	double cooperationParteB;
	double perturbationParteB;

	double delta;
	vector<double> minPos, maxPos, minVel, maxVel;
};
