#include ".\particleHW.h"

//=================================
// destructor
HWparticle::~HWparticle(void){}

//=================================
// constructor, default
HWparticle::HWparticle(void){}

//=================================
// constructor for new particles of Part A
HWparticle::HWparticle(
	const fvec &variaveisProblema,  
	const vec &typeperiodEntry, 
	const vec &windgenEntry, 
	vec xprefinalEntry, 
	vec xfinalEntry,
	const vec &capmin, 
	const vec &capmax, 
	const vec &inflow, 
	const vec &numres, 
	const vec &qmin, 
	const vec &qmax, 
	boost::mt19937 &rng, 
	HWfitnessfunction *pFf, 
	const double &CpG, 
	const double &TauG,
	ProblemSelection ProblemSelect)
{

	
	vec xprefinal = xprefinalEntry;
	vec xfinal = xfinalEntry;

	unsigned int numR = variaveisProblema (12);
	unsigned int DimIn = numR*variaveisProblema(13);

	//==============================
	double maxPos, minPos;
	vec xanterior (DimIn); vec z (DimIn); 
	
	
	for (unsigned int i=0; i<DimIn; i++)
	{
		xanterior[i]=xfinal[i];
		
		//gama
		double gama = HYB_func::Gama_func(i, numres, xfinal, capmin, ProblemSelect);

		//construo de sumomega 
		double sumomega = HYB_func::sumOmega_func(i, numres, z, Pos, ProblemSelect);

		//minPos e maxPos
		HYB_func::MinMaxPos	(
			maxPos,minPos,
			i, 
			xanterior,
			Pos,
			capmin,
			capmax,
			inflow, 
			numres, 
			qmin, 
			qmax, 
			sumomega,
			gama, 
			ProblemSelect);
		
		//=============================
		//no caso de os limites se interceptarem, assume-se que no turbina nem bomba nada
		//como a variavel minPos tem de ser distinta da maxPos, porque as funoes de aleatorios do boost nao aceitam valores iguais
		//adoptam-se numeros muito pequenos, e distintos, cujo impact  residual (nada significativo)
		if ( minPos > maxPos)
		{
			minPos=0.001;
			maxPos=0.002;
		}

		//==============================
		//generates positions with random numbers
		boost::uniform_real<double> uniform_pos(minPos, maxPos);  
		boost::variate_generator<boost::mt19937&, boost::uniform_real<double> >  uniform_sampler_position(rng, uniform_pos);
		Pos.push_back(uniform_sampler_position());
		
		//xprefinal
		xprefinal[i+numR]=xanterior[i]+inflow[i]+sumomega - Pos[i];

		//obriga a respeitar a capmin
		if (xprefinal[i+numR]<capmin[i])
		{
			Pos[i]-=capmin[i]-xprefinal[i+numR];
			xprefinal[i+numR]=xanterior[i]+inflow[i]+sumomega - Pos[i];
		}

		//vetor z
		z[i]=max(0.0,(xprefinal[i+numR]-capmax[i])); 

		//xfinal
		xfinal[i+numR]=xprefinal[i+numR]-z[i];

		//==============================
		//generates velocities with random numbers
		boost::uniform_real<double> uniform_vel(qmin[i], qmax[i]);
		boost::variate_generator<boost::mt19937&, boost::uniform_real<double> >  uniform_sampler_velocities(rng, uniform_vel); 
		Vel.push_back(uniform_sampler_velocities());
	}
	
	//==============================
	//generates weights with random numbers (between 0 and 1)
	boost::uniform_real<double> uniform_weight(0,1);
	boost::variate_generator<boost::mt19937&, boost::uniform_real<double> >  uniform_sampler_weights(rng, uniform_weight); 

	inertia=uniform_sampler_weights();
	memory=uniform_sampler_weights();
	cooperation=uniform_sampler_weights();
	perturbation=uniform_sampler_weights(); 
	
	//==============================
	this->ff = pFf;
	
	myBestPos=Pos;
	globalBestPos=Pos;	
	
	this->HWUpdateFitness(
		variaveisProblema,  
		typeperiodEntry, 
		windgenEntry, xanterior, xfinal);

	myBestFitness=Fitness;
}

//=================================
// constructor for particles of Part C (position and velocity known in advance)
HWparticle::HWparticle(
	const fvec &variaveisProblema,  
	const vec &typeperiodEntry, 
	const vec &windgenEntry, 
	vec xprefinalEntry, 
	vec xfinalEntry,
	vector<double> tempPos,
	vector<double> tempBeforePos,
	vector<double> tempBestPos,
	const mat &W2, const rowvec &B2, const rowvec &Parameter1, const rowvec &Parameter2,
	const float &actFcnFlatness,
	const ActivationFunctions &NeuronActFunct2,
	const NormalizationMethod &NormMethod,
	const vec &capmin, 
	const vec &capmax, 
	const vec &inflow, 
	const vec &numres, 
	const vec &qmin, const vec &qmax, 
	boost::mt19937 &rng, 
	double Inertia, double Memory, double Cooperation, double Perturbation,
	HWfitnessfunction *pFf, 
	const double &CpG,
	ProblemSelection ProblemSelect,
	const Initi_Vel &Vel_init)
{
	this->inertia=Inertia;
	this->memory=Memory;
	this->cooperation=Cooperation;
	this->perturbation=Perturbation;
	
	unsigned int numR = variaveisProblema (12);
	unsigned int DimIn = numR*variaveisProblema(13);
	unsigned int DimHidden = variaveisProblema(14);
	
	
	//==============================
	//extrair as particulas
	unsigned short numPatterns =3;
	mat To_expand_Pos (numPatterns, DimIn);
	mat Com_Pos (numPatterns, DimHidden);

	for (unsigned int j=0; j<DimHidden; j++)
	{
		Com_Pos(0,j)=tempPos[j];
		Com_Pos(1,j)=tempBestPos[j];
		Com_Pos(2,j)=tempBeforePos[j];
	}
	
	//==============================
	//expandir as particulas
	General_functions::transitionLayer (
		To_expand_Pos, //para ser actualizado
		numPatterns, //os dois vetores a serem expandidos
		DimHidden, 
		DimIn, 
		Com_Pos, 
		W2, 
		B2, 
		actFcnFlatness, 
		NeuronActFunct2);
	

	//==============================
	//desnormalizar as particulas
	mat DesNormPositions (numPatterns, DimIn);

	DesNormPositions =  Normalization::DES_Normalization_function (
		Parameter1, 
		Parameter2, 
		To_expand_Pos, 
		numPatterns, 
		DimIn, 
		NormMethod);
	
		
	expGlobalBestPos.resize(DimIn);
	expPos.resize(DimIn); 
	for (unsigned int j=0; j<DimIn; j++)
	{
		expPos[j]=DesNormPositions(0 ,j); 
		expGlobalBestPos[j]=DesNormPositions(1,j);

		switch (Vel_init)
		{ 
			case RANDOM_VEL: 
			{
				//define distribution for velocities according to [minVel, maxVel] for i-th dimension
				boost::uniform_real<double> uniform_vel(qmin[j], qmax[j]);
				boost::variate_generator<boost::mt19937&, boost::uniform_real<double> >  uniform_sampler_velocities(rng, uniform_vel); 
				Vel.push_back(uniform_sampler_velocities());
				break;
			}
			case TRANSFER_VEL: 
			{
				//velocities are initialize as the diference V(k) = X(k)-X(k-1) 
				this->Vel.push_back(DesNormPositions(0,j)-DesNormPositions(2,j));
				break;
			}
			default: cout << "ERROR!!! ";
		}
	}

	//------------------COLOCAR O PRIMEIRO ENXAME DENTRO DO DOMNIO~!!!!
	vec xprefinal = xprefinalEntry;
	vec xfinal = xfinalEntry ;


	vec xanterior (DimIn);
	double maxPos; double minPos; 
	vec z (DimIn);

	for (unsigned int i=0; i<DimIn; i++)
	{
		xanterior[i]=xfinal[i];
		
		//gama
		double gama = HYB_func::Gama_func(i, numres, xfinal, capmin, ProblemSelect);

		//construo de sumomega 
		double sumomega = HYB_func::sumOmega_func(i, numres, z, expPos, ProblemSelect);

		//minPos e maxPos
		HYB_func::MinMaxPos	(
			maxPos,minPos,
			i, 
			xanterior,
			expPos,
			capmin, capmax,
			inflow, 
			numres, 
			qmin, qmax, 
			sumomega,
			gama, 
			ProblemSelect);
	

		//os limites do domnio funcionam como um espelho!!!! ou apenas coloco inrcia a zero
#if 1
		if (expPos[i]<minPos)
		{
			expPos[i]=minPos;
		}
		//Vel[i]=-Vel[i];}
		//inertia=0.0;}

		if (expPos[i]>maxPos)
		{
			expPos[i]=maxPos;
		}
		//Vel[i]=-Vel[i];}
		//inertia=0.0;}
#endif
		//xprefinal
		xprefinal[i+numR]=xanterior[i]+inflow[i]+sumomega - expPos[i];
		
		//obriga a respeitar a capmin
		if (xprefinal[i+numR]<capmin[i])
		{
			expPos[i]-=capmin[i]-xprefinal[i+numR];
			xprefinal[i+numR]=xanterior[i]+inflow[i]+sumomega - expPos[i];
		}
		
		//vetor z
		z[i]=max(0.0,(xprefinal[i+numR]-capmax[i])); 

		//xfinal
		xfinal[i+numR]=xprefinal[i+numR]-z[i];
	}

	for (unsigned int i=0; i<DimIn; i++)
    {
        this->Pos.push_back(expPos[i]);
        this->myBestPos.push_back(expPos[i]);
		this->globalBestPos.push_back(expGlobalBestPos[i]);
    }

	this->Vel = Vel;
	this->ff = pFf;
	
	
	this->HWUpdateFitness(
		variaveisProblema,  
		typeperiodEntry, 
		windgenEntry, 
		xanterior, 
		xfinal);

	myBestFitness=Fitness;

}

//=================================
// gets the current fitness value but does not recalculate fitness each time
double HWparticle::HWGetFitness()
{
	return Fitness;
}

//=================================
// update the fitness
void HWparticle::HWUpdateFitness(
	const fvec &variaveisProblema,  
	const vec &typeperiodEntry, 
	const vec &windgenEntry,
	vec xanterior,
	vec xfinal)
{
	this->Fitness=this->ff->calcFitness(variaveisProblema, typeperiodEntry, windgenEntry, Pos, xanterior, xfinal);
}

//=================================
// print function
void HWparticle::HWPrint()
{
	cout << "=============================" << endl;	
	cout << " x1 : " << Pos[0] << " x2 : " << Pos[1] << " f: " << Fitness << endl;
	cout << " v1 : " << Vel[0] << " v2 : " << Vel[1] << endl;
	cout <<	" gB1: " << globalBestPos[0]<< " gB2: " << globalBestPos[1] << endl;
	cout << " in : " << inertia << " me : " << memory << " co : " << cooperation;
	//cout << " pe : " << perturbation << endl << endl;
	cout << endl;
}

//=================================
//get the particle's own previous best position (pBest)
vector <double > HWparticle::HWGetBestPos()
{
	return myBestPos;
}

//=================================
//set global best position gBest for this particle
void HWparticle::HWSetBestPos(vector<double> newGlobalBestPos)
{
	globalBestPos=newGlobalBestPos;
}

//=================================
//get particle's current position (location, vector X)
vector <double > HWparticle::HWGetPos()
{
	return Pos;
}

//=================================
//get particle's current velocity
vector <double > HWparticle::HWGetVel()
{
	return Vel;
}

//=================================
// mutate the internal parameters (EPSO)
//rng is parameter so that lognormal mutation can be called
void HWparticle::HWMutate(
	boost::mt19937 &rng, 
	const double &TauG)
{
	boost::normal_distribution<double> norm_dist(0.0, 1.0);
	boost::variate_generator<boost::mt19937&, boost::normal_distribution<double> >  rng_sampler(rng, norm_dist);
	
	double ranNum;
	ranNum=rng_sampler();
	//do {ranNum=rng_sampler();}
	//while (TauG*ranNum<-0.95);

	//inertia=inertia * pow(exp(ranNum), TauG); //lognormal type 
#if 1
	inertia=inertia + ranNum*TauG; // gaussian type
	if (inertia<0) inertia=0;
	if (inertia>1) inertia=1;

	ranNum=rng_sampler();
	//memory=memory * pow(exp(ranNum), TauG);
	memory=memory + ranNum*TauG;
	if (memory<0) memory=0;
	if (memory>1) memory=1;

	ranNum=rng_sampler();
	//cooperation=cooperation * pow(exp(ranNum), TauG);
	cooperation=cooperation + ranNum * TauG;
	if (cooperation<0) cooperation=0;
	if (cooperation>1) cooperation=1;
	
	ranNum=rng_sampler();
	//perturbation=perturbation * pow(exp(ranNum), TauG);
	perturbation=perturbation + ranNum*TauG;
	if (perturbation<0) perturbation=0;
	if (perturbation>1) perturbation=1;
#endif

#if 0 // w=w[1+tauN(0,1)] version
	inertia=inertia + inertia*ranNum*TauG; 
	//if (inertia<0) inertia=0;
	//if (inertia>1) inertia=10;

	ranNum=rng_sampler();
	//memory=memory * pow(exp(ranNum), TauG);
	memory=memory + memory*ranNum*TauG;
	//if (memory<0) memory=0;
	//if (memory>1) memory=10;

	ranNum=rng_sampler();
	//cooperation=cooperation * pow(exp(ranNum), TauG);
	cooperation=cooperation + cooperation*ranNum * TauG;
	//if (cooperation<0) cooperation=0;
	//if (cooperation>1) cooperation=5;
	
	ranNum=rng_sampler();
	//perturbation=perturbation * pow(exp(ranNum), TauG);
	perturbation=perturbation + perturbation*ranNum*TauG;
	//if (perturbation<0) perturbation=0;
	//if (perturbation>1) perturbation=5;
#endif
}

//=================================
//mutao que consiste em reinicializar os pesos das particulas de forma aleatoria
void HWparticle::RandomHWMutate(boost::mt19937 &rng)
{
   
    //define distribution for weights
    boost::uniform_real<double> uniform_weight(0,5);
    boost::variate_generator<boost::mt19937&, boost::uniform_real<double> >  uniform_sampler_weights(rng, uniform_weight);

    inertia=uniform_sampler_weights();
    memory=uniform_sampler_weights();
    cooperation=uniform_sampler_weights();
    perturbation=uniform_sampler_weights();
}

//=================================
// perform the movement of particles according to EPSO rules
void HWparticle::Movement(
	boost::mt19937& rng,
	const fvec &variaveisProblema,   
	const vec &typeperiodEntry, 
	const vec &windgenEntry, 
	vec xprefinalEntry, 
	vec xfinalEntry,
	const vec &capmin, 
	const vec &capmax, 
	const vec &inflow, 
	const vec &numres, 
	const vec &qmin, 
	const vec &qmax,
	const double &CpG,
	ProblemSelection ProblemSelect,
	const Initi_Vel &Vel_init,
	const unsigned int &number, //numero da iteraao relativa a parte A e parte C
	const unsigned int &maxIterI, 
	const unsigned int &maxIterS)
{
	
	double inertiaTerm, memoryTerm, cooperationTerm; // perturbationTerm;
	double tempVel;
	double tempPos;
		
	boost::uniform_real<double> uniform_dist(0,1);
	boost::variate_generator<boost::mt19937&, boost::uniform_real<double> >  uniform_sampler(rng, uniform_dist);

	double ranNum;
	double disturbator;
	disturbator = uniform_sampler(); // for disturbing the global best of the particular particle
	
	vec xprefinal = xprefinalEntry;
	vec xfinal = xfinalEntry;

	unsigned int numR = variaveisProblema (12);
	unsigned int DimIn = numR*variaveisProblema(13);
	
	vec xanterior (DimIn);
	double maxPos, minPos; 
	vec z (DimIn);

	for (unsigned int i=0; i<DimIn; i++)
	{
		inertiaTerm = inertia * Vel[i]; // iner * v(t-1)
		memoryTerm = memory * ( myBestPos[i] - Pos[i] ); // mem * (myBest - pos)
			
		if (uniform_sampler()<CpG) // per-dimension stochastic star - default for EPSO
		{
			ranNum = disturbator;
			globalBestPos[i]=globalBestPos[i] * (1 + perturbation * ranNum);
			cooperationTerm = cooperation * (globalBestPos[i] - Pos[i]);
		}
		
		else
		{	
			cooperationTerm = 0;
		}

		//========================
		//this "if" condition is to be activated only for normal EPSO
		//it aims to add random initializations (with normal EPSO) of velocity for the moments corresponding to the space transitions in Hybrid
		//this "if" condition is to be activated only for normal EPSO at iterations: maxIterI (first moment of partB) and 
		//maxIterI+maxIterS (first moment of partC)
		if( (number == maxIterI+1) || (number == (maxIterI+maxIterS+1))) //nota
		{
			switch (Vel_init)
			{ 
				case RANDOM_VEL: 
				{
					//random
					boost::uniform_real<double> uniform_vel(qmin[i], qmax[i]);
					boost::variate_generator<boost::mt19937&, boost::uniform_real<double> >  uniform_sampler_velocities(rng, uniform_vel); 
					tempVel = uniform_sampler_velocities();	
					break;
				}
				case TRANSFER_VEL: 
				{
					//velocities are initialize as the diference V(k) = X(k)-X(k-1) 
					//this corresponds to the normal evolution / transference of velocities
					tempVel = inertiaTerm + memoryTerm + cooperationTerm; // v(t)
					break;
				}
				default: cout << "ERROR!!! ";
			}
		}
		else
		{
			//this corresponds to the normal evolution / transference of velocities
			// v(t)
			tempVel = inertiaTerm + memoryTerm + cooperationTerm; 
		}
		
		if (tempVel<qmin[i]) {tempVel=qmin[i];}
		if (tempVel>qmax[i]) {tempVel=qmax[i];}

		// update velocity
		Vel[i]=tempVel; 

		// x(t+1) = x(t) + v(t)
		tempPos=Pos[i]+Vel[i]; 

		xanterior[i]=xfinal[i];
		
		//gama :: concerns the set of reservoirs imediatly downstream, to further ensure the min lim of each reservoir
		double gama = HYB_func::Gama_func(i, numres, xfinal, capmin, ProblemSelect);

		//construo de sumomega :: the set of reservoirs imediately upstream from the one in analysis
		double sumomega = HYB_func::sumOmega_func(i, numres, z, Pos, ProblemSelect);

		//dynamic limits for water volumes: minPos & maxPos
		//minPos e maxPos
		HYB_func::MinMaxPos	(
			maxPos,minPos,
			i, 
			xanterior,
			Pos,
			capmin,
			capmax,
			inflow, 
			numres, 
			qmin, 
			qmax, 
			sumomega,
			gama, ProblemSelect);
		
		//=================================
		//os limites do domnio funcionam como um espelho
		if (tempPos<minPos)
		{
			tempPos=minPos;
			Vel[i]=-Vel[i];
		}
		
		if (tempPos>maxPos)
		{
			tempPos=maxPos;
			Vel[i]=-Vel[i];
		}
		
		
		//=================================
		// update position
		Pos[i]=tempPos; 
		
		//=================================
		// update vector "xprefinal"
		xprefinal[i+numR]=xanterior[i]+inflow[i]+sumomega - Pos[i];

		//=================================
		//ensures limit capmin
		if (xprefinal[i+numR]<capmin[i])
		{Pos[i]-=capmin[i]-xprefinal[i+numR];
		xprefinal[i+numR]=xanterior[i]+inflow[i]+sumomega - Pos[i];}

		//=================================
		//calculus of vetor z (the spilled water volume)
		z[i]=max(0.0,(xprefinal[i+numR]-capmax[i])); 

		//=================================
		//calculus of xfinal
		xfinal[i+numR]=xprefinal[i+numR]-z[i];
		
		//=================================
		//ensures velocities limits
		if (Vel[i]<qmin[i]) {Vel[i]=qmin[i];} // in case of asymmetric velocity limits
		if (Vel[i]>qmax[i]) {Vel[i]=qmax[i];}
	}
	
	this->HWUpdateFitness(
		variaveisProblema,  
		typeperiodEntry, 
		windgenEntry, 
		xanterior,
		xfinal);

	if (ff->HWfitnessCompare(this->Fitness, this->myBestFitness)) //if (this->Fitness > this->myBestFitness)
	{	
		this->myBestFitness = this->Fitness;
		this->myBestPos = this->Pos;
	}
}


//=================================
//get epso parameters
double HWparticle::GetInertia()
{
	return inertia;
}

double HWparticle::GetMemory()
{
	return memory;
}

double HWparticle::GetCooperation()
{
	return cooperation;
}

double HWparticle::GetPerturbation()
{
	return perturbation;
}


