#include "swarmHW.h"


//=================================
// destructor
HWswarm::~HWswarm(void){}

//=================================
// constructor of swarm for PartA
HWswarm::HWswarm(
	boost::mt19937 &rng, 
	const unsigned int &numParticles, 
	const fvec &variaveisProblema,  
	const vec &typeperiodEntry, 
	const vec &windgenEntry, 
	vec xprefinalEntry, 
	vec xfinalEntry,
	const vec &capmin, 
	const vec &capmax, 
	const vec &inflow, 
	const vec &numres, 
	const vec &qmin, 
	const vec &qmax, 
	bool newMinimize, 
	HWfitnessfunction &newFf, 
	const double &CpG,  
	const double &TauG,
	ProblemSelection ProblemSelect) : ff(newFf) 
{
	this->numEvalsParteA = 0;
	this->ff = newFf; 
	this->minimize = newMinimize;

	ff.HWSetMinimize(minimize);

	for (unsigned int i=0; i<numParticles; i++) // initialize list of particles
	{
		EnxameParteA.push_back(
			HWparticle(
				variaveisProblema,  
				typeperiodEntry, 
				windgenEntry , 
				xprefinalEntry, 
				xfinalEntry,
				capmin, 
				capmax, 
				inflow, 
				numres, 
				qmin, 
				qmax, 
				rng, 
				&ff, 
				CpG, 
				TauG,
				ProblemSelect));
	}

	
	EnxameParteA.sort(ff.particleCompare); // may be stupid: sort list of particles according to their fitness

	this->gBestFitness = EnxameParteA.begin()->HWGetFitness();
	this->gBestPos = EnxameParteA.begin()->HWGetPos();  // get the global best for the beginning

}


//=================================
// constructor of swarm for PartC
HWswarm::HWswarm(
	boost::mt19937 &rng,
	list <particle> &lastSegundoSwarm,
	list <particle> &beforeLastSegundoSwarm,
	const unsigned int &numParticles,
	const fvec &variaveisProblema,   
	const vec &typeperiodEntry, 
	const vec &windgenEntry, 
	vec xprefinalEntry, 
	vec xfinalEntry,
	const vec &capmin, 
	const vec &capmax, 
	const vec &inflow, 
	const vec &numres, 
	const vec &qmin, 
	const vec &qmax, 
	const mat &W2,
	const rowvec &B2, const rowvec &Parameter1, const rowvec &Parameter2,
	const float &actFcnFlatness,
	const ActivationFunctions &NeuronActFunct2,
	const NormalizationMethod &NormMethod,
	bool newMinimize, 
	HWfitnessfunction &newFf, 
	const double &CpG, 
	const double &TauG,
	ProblemSelection ProblemSelect,
	const Initi_Vel &Vel_init):ff(newFf)
{
	this->numEvalsParteC = 0;
	this->minimize = newMinimize;
	this->ff = newFf;	
	
	unsigned int DimHidden = variaveisProblema (14);
	ff.HWSetMinimize(minimize);  
	
	
	list<particle>::iterator pParticle;
	list<particle>::iterator beforeParticle;
	pParticle=lastSegundoSwarm.begin();
	beforeParticle=beforeLastSegundoSwarm.begin();

	while(pParticle!=lastSegundoSwarm.end())
	{
		vector<double> tempPos, tempBeforePos, tempBestPos;
		for (unsigned int i=0; i<DimHidden; i++)
		{
			tempPos.push_back(pParticle->GetPos()[i]);
			tempBestPos.push_back(pParticle->GetBestPos()[i]);

			tempBeforePos.push_back(beforeParticle->GetVel()[i]);
		}

		Fitness=pParticle->GetFitness();

		Inertia=pParticle->GetSInertia();
		Memory=pParticle->GetSMemory();
		Cooperation=pParticle->GetSCooperation();
		Perturbation=pParticle->GetSPerturbation();

		EnxameParteC.push_back(HWparticle(
			variaveisProblema,  
			typeperiodEntry, 
			windgenEntry, 
			xprefinalEntry, 
			xfinalEntry,
			tempPos,
			tempBeforePos, 
			tempBestPos,
			W2,
			B2, Parameter1, Parameter2,
			actFcnFlatness,
			NeuronActFunct2,
			NormMethod,
			capmin, 
			capmax, 
			inflow, 
			numres, 
			qmin, 
			qmax, 
			rng,
			Inertia,
			Memory, 
			Cooperation, 
			Perturbation, 
			&ff, 
			CpG,
			ProblemSelect,
			Vel_init));

		pParticle++;
		beforeParticle++;
	}
		
	EnxameParteC.sort(ff.particleCompare); // may be stupid: sort list of particles according to their fitness

	this->gBestFitness = EnxameParteC.begin()->HWGetFitness();
	this->gBestPos = EnxameParteC.begin()->HWGetPos();  // get the global best for the beginning
}


void HWswarm::NextIterationPrimeiro(
	boost::mt19937 &rng, 
	const unsigned int &numberPrimeiro, //numero da iteraao relativa a parte A
	const unsigned int &maxIterI, 
	const unsigned int &maxIterS,
	const unsigned int &bigMaxIterI,
	const fvec &variaveisProblema,   
	const vec &typeperiodEntry, 
	const vec &windgenEntry, 
	vec xprefinalEntry, 
	vec xfinalEntry,
	const vec &capmin, 
	const vec &capmax, 
	const vec &inflow, 
	const vec &numres, 
	const vec &qmin, 
	const vec &qmax, 
	const unsigned int &trainDim, //ok 
	const unsigned int &TBegin, //ok 
	const unsigned int &TEnd,//ok 
	const unsigned int &testDim,//ok 
	const double &CpG, //ok
	const double &TauG,
	ProblemSelection ProblemSelect,
	const Initi_Vel &Vel_init) //ok
{
	
	HWparticle tempParticle;

	list<HWparticle>::iterator pParticle;
	pParticle=EnxameParteA.begin();	// this is the best particle

	if (ff.HWfitnessCompare(pParticle->HWGetFitness(), gBestFitness))
	{
		gBestPos = pParticle->HWGetPos(); // new gBest found - update.
		gBestFitness = pParticle->HWGetFitness(); 
	}
	

	while(pParticle!=EnxameParteA.end())
	{
		pParticle->HWSetBestPos(gBestPos); //new gBest set
		tempParticle=*pParticle; // replicate particle, make a copy
		tempParticle.HWMutate(rng, TauG); //  mutate internal parameters

		tempParticle.Movement(
			rng, 
			variaveisProblema, 
			typeperiodEntry, 
			windgenEntry, 
			xprefinalEntry, 
			xfinalEntry, 
			capmin, capmax, 
			inflow, 
			numres, 
			qmin, qmax, 
			CpG, 
			ProblemSelect, 
			Vel_init,
			numberPrimeiro,
			maxIterI,
			maxIterS); // move the replicated particle

		(*pParticle).Movement(
			rng, 
			variaveisProblema, 
			typeperiodEntry, 
			windgenEntry, 
			xprefinalEntry, 
			xfinalEntry, 
			capmin, capmax, 
			inflow,
			numres, 
			qmin, qmax,
			CpG, 
			ProblemSelect, 
			Vel_init,
			numberPrimeiro,
			maxIterI,
			maxIterS); // move the original particle

		numEvalsParteA++;
		numEvalsParteA++;

		if ( ff.particleCompare(*pParticle, tempParticle) )
		{ // original particle chosen - nothing to do
		}
		else
		{ // mutated particle chosen
			(*pParticle)=tempParticle;
		}	
		pParticle++; // iterate through all the particles
	}

	EnxameParteA.sort(ff.particleCompare); // order the list of particles
	
	if (numberPrimeiro==(maxIterI)) //nota
	{
		pParticle=EnxameParteA.begin();
		while(pParticle!=EnxameParteA.end())
		{
			lastPrimeiroSwarm.push_back(*pParticle);
			pParticle++;
		}
	}

	//======================================
	//to store the before Pos to transfer velocity from S to S'
	if (numberPrimeiro==(maxIterI-1)) //nota
	{
		pParticle=EnxameParteA.begin();
		while(pParticle!=EnxameParteA.end())
		{
			beforeLastPrimeiroSwarm.push_back(*pParticle);
			pParticle++;
		}
	}
	// iteration finished
}


void HWswarm::NextIterationTerceiro(
	boost::mt19937& rng, //ok
	const unsigned int &numberTerceiro, //ok
	const unsigned int &maxIterF, 
	const fvec &variaveisProblema,   
	const vec &typeperiodEntry, 
	const vec &windgenEntry,
	vec xprefinalEntry, 
	vec xfinalEntry,
	const vec &capmin, 
	const vec &capmax, 
	const vec &inflow, 
	const vec &numres, 
	const vec &qmin, 
	const vec &qmax, 
	const double &CpG,//ok
	const double &TauG,
	ProblemSelection ProblemSelect,
	const Initi_Vel &Vel_init)//ok
{
	HWparticle tempParticle;  

	list<HWparticle>::iterator pParticle;
	pParticle=EnxameParteC.begin();	// this is the best particle

	if (ff.HWfitnessCompare(pParticle->HWGetFitness(), gBestFitness))
	{
		gBestPos = pParticle->HWGetPos(); // new gBest found - update.
		gBestFitness = pParticle->HWGetFitness(); 
	}
	//se o bloco #if seguinte estiner activo, faz a reinicializao dos pesos
#if 0
	int contador=0;

	while(pParticle!=EnxameParteC.end())
	{
		pParticle->HWSetBestPos(gBestPos); // new gBest set
		tempParticle=*pParticle; // replicate particle, make a copy

		if (((i%1000)==0) /*&& (((contador+1)%4)==0)*/)  //  1/4 das particulas
		//if (((i%1000)==0) && (((contador+1)%2)==0))  // 1/2 das particulas
		//if (((i%1000)==0) && ((contador%4)!=0))  // 3/4 das particulas
		{
			//cout<<"i" << i<< " contador " << contador << endl;
			tempParticle.RandomHWMutate(rng); //  tau = 0.2, mutate internal parameters
			contador=contador++;
		}
		else
		{
			//cout<<"i" <<i<< " contador " << contador << endl;
			tempParticle.HWMutate(rng, TauG); //  tau = 0.2, mutate internal parameters
			contador=contador++;
		}
	}
#endif


	while(pParticle!=EnxameParteC.end())
	{
		pParticle->HWSetBestPos(gBestPos); // new gBest set
		tempParticle=*pParticle; // replicate particle, make a copy
		tempParticle.HWMutate(rng, TauG); //  tau = 0.2, mutate internal parameters

		tempParticle.Movement(
			rng, 
			variaveisProblema,
			typeperiodEntry, 
			windgenEntry, 
			xprefinalEntry, 
			xfinalEntry, 
			capmin, capmax, 
			inflow, 
			numres, 
			qmin, qmax, 
			CpG, 
			ProblemSelect, 
			Vel_init,
			numberTerceiro,
			maxIterF+1000, //here the maxIterF is added by 1000 to ensure that the if condition inside the movement function is never activated in partC
			maxIterF+1000); // here the same (the if condition in movement fucntion is to be activated only for part A) 

		(*pParticle).Movement(
			rng, 
			variaveisProblema,
			typeperiodEntry, 
			windgenEntry, 
			xprefinalEntry, 
			xfinalEntry,  
			capmin, capmax, 
			inflow, 
			numres, 
			qmin, qmax, 
			CpG, 
			ProblemSelect, 
			Vel_init,
			numberTerceiro,
			maxIterF+1000,
			maxIterF+1000); // move the original particle

		numEvalsParteC++;
		numEvalsParteC++;

		if ( ff.particleCompare(*pParticle, tempParticle) )
		{ // original particle chosen - nothing to do
		}
		else
		{ // mutated particle chosen
			(*pParticle)=tempParticle;
		}	
		pParticle++; // iterate through all the particles
	}

	EnxameParteC.sort(ff.particleCompare); // order the list of particles
}

unsigned int HWswarm::HWGetNumEvalsParteA()
{
	return this->numEvalsParteA;
}

unsigned int HWswarm::HWGetNumEvalsParteC()
{
	return this->numEvalsParteC;
}

double HWswarm::HWGetBestFitness()
{
	return gBestFitness;
}

vector<double> HWswarm::HWGetBestPos()
{
	return gBestPos;
}

