//#ifndef SWARMHW_H
//#define SWARMHW_H

//=================================
// include guard
#pragma once

//=================================
// included dependencies
#include <iostream>
#include <fstream>
#include <iomanip>
#include <ctime>
#include <list>
#include <vector>
#include <armadillo>

#include <boost/random.hpp>

#include "particleHW.h"
#include "fitnessfunctionHW.h"

#include "../EPSO/swarm.h"
#include "../EPSO/particle.h"
#include "..\enumerations.h"

//=================================
// namespaces
using namespace std;
using namespace arma;

//=================================
// the HWswarm class
class HWswarm
{
public:
	//=================================
	// destructor
	~HWswarm(void); 

	//=================================
	// constructor of swarm for PartA
	HWswarm
	(
		boost::mt19937& rng, 
		const unsigned int &numParticles, 
		const fvec &variaveisProblema,  
		const vec &typeperiodEntry, 
		const vec &windgenEntry, 
		vec xprefinalEntry, 
		vec xfinalEntry,
		const vec &capmin, 
		const vec &capmax, 
		const vec &inflow, 
		const vec &numres, 
		const vec &qmin, 
		const vec &qmax, 
		bool newMinimize,
		HWfitnessfunction &ff, 
		const double &CpG, 
		const double &TauG,
		ProblemSelection ProblemSelect
	); 

	//=================================
	// constructor of swarm for PartC
	HWswarm
	(
		boost::mt19937 &rng,
		list<particle> &lastSegundoSwarm,
		list <particle> &beforeLastSegundoSwarm,
		const unsigned int &numParticles, 
		const fvec &variaveisProblema,   
		const vec &typeperiodEntry, 
		const vec &windgenEntry, 
		vec xprefinalEntry, 
		vec xfinalEntry,
		const vec &capmin, 
		const vec &capmax, 
		const vec &inflow, 
		const vec &numres, 
		const vec &qmin, 
		const vec &qmax, 
		const mat &W2,
		const rowvec &B2, const rowvec &Parameter1, const rowvec &Parameter2,
		const float &actFcnFlatness,
		const ActivationFunctions &NeuronActFunct2,
		const NormalizationMethod &NormMethod,
		bool minimize, 
		HWfitnessfunction &ff, 
		const double &CpG, 
		const double &TauG,
		ProblemSelection ProblemSelect,
		const Initi_Vel &Vel_init
	);
	
	//=================================
	void NextIterationPrimeiro(
		boost::mt19937& rng,
		const unsigned int &numberPrimeiro, 
		const unsigned int &maxIterI, 
		const unsigned int &maxIterS, 
		const unsigned int &bigMaxIterI,
		const fvec &variaveisProblema,   
		const vec &typeperiodEntry, 
		const vec &windgenEntry,
		vec xprefinalEntry, 
		vec xfinalEntry,
		const vec &capmin, 
		const vec &capmax, 
		const vec &inflow, 
		const vec &numres, 
		const vec &qmin, 
		const vec &qmax, 
		const unsigned int &trainDim, //ok 
		const unsigned int &TBegin, //ok 
		const unsigned int &TEnd,//ok 
		const unsigned int &testDim,//ok 
		const double &CpG,
		const double &TauG,
		ProblemSelection ProblemSelect,
		const Initi_Vel &Vel_init);

	//=================================
	void NextIterationTerceiro(
		boost::mt19937& rng,
		const unsigned int &numberTerceiro, 
		const unsigned int &maxIterF, 
		const fvec &variaveisProblema,   
		const vec &typeperiodEntry, 
		const vec &windgenEntry, 
		vec xprefinalEntry, 
		vec xfinalEntry,
		const vec &capmin, 
		const vec &capmax, 
		const vec &inflow, 
		const vec &numres, 
		const vec &qmin, 
		const vec &qmax, 
		const double &CpG,
		const double &TauG,
		ProblemSelection ProblemSelect,
		const Initi_Vel &Vel_init);
	
	

	unsigned int HWGetNumEvalsParteA();
	unsigned int HWGetNumEvalsParteC();

	double HWGetBestFitness();
	vector<double> HWGetBestPos();
	
	
	list<HWparticle> EnxameParteA;	
	list<HWparticle> EnxameParteC;
	
	list<HWparticle> lastPrimeiroSwarm;//list holding the last swarm for the neural net
	list<HWparticle> lastTerceiroSwarm;

	//======================================
	//to transfer velocity to reduced space as V(k) = X(k) - X(k-1)
	list<HWparticle> beforeLastPrimeiroSwarm; 


private:

	double gBestFitness; // global best's fitness
	vector<double > gBestPos; // global best position
	

	unsigned int numEvalsParteA; // number of fitness function evaluations performed
	unsigned int numEvalsParteC;
	HWfitnessfunction &ff;
	
	bool minimize; // minimizing or maximizing
	
	double Fitness;//ok

	double Inertia;//ok
	double Memory;//ok
	double Cooperation;//ok
	double Perturbation;//ok
};
