//=================================
// include guard
#pragma once

//=================================
// included dependencies
#include <iostream>
#include <math.h>
#include <stdio.h>
#include <vector>
#include <stdlib.h>
#include <armadillo>

#include "ITL_networks\General_Functions.h"
#include "..\enumerations.h"
//=================================
// namespaces
using namespace std;
using namespace arma;


//=================================
void Ent_Weights1Bias1_function (
	const unsigned short &DimHidden, 
	const unsigned short &DimIn, 
	const unsigned short &TrainDim, 
	const mat &HiddenTrain, 
	const mat &TrainSet, 
	const float &actFcnFlatness, 
	const double &sig2,
	const float &momRate, 
	const float &lrCoeff1, 
	mat &Weights1, //para ser alterado directamente 
	const mat &prev_Weights1, 
	rowvec &Bias1, //para ser alterado directamente
	const rowvec &prevBias1, 
	const ActivationFunctions &NeuronActFunct)
{
	
	//##### BackForward Step == MaxE Criterium #####
	rowvec y_i(DimHidden), y_j(DimHidden), dij(DimHidden);
	double Vij_out;
							
	double dTrainDim = static_cast<double>(TrainDim);

	for (unsigned short i=0; i<TrainDim; i++) //loop over pattern i (Cross samples)
	{
		for (unsigned short j=0; j<TrainDim; j++) //loop over pattern j (Cross samples)
		{
			y_i = HiddenTrain(i,span()); //pattern i extraction 
			y_j = HiddenTrain(j,span()); //pattern j extraction
			dij = y_i - y_j; // the difference bettween patterns i and j
			
			Vij_out = General_functions::Vij_function(dij, DimHidden, sig2); // Vij = G(yi-yj, 2sig^2)

			// dV/Dw = dV/dyi * dyi/dw
			mat delta_weights1(DimIn,DimHidden);
			for(unsigned short hid=0; hid<DimHidden; hid++) //for each output neuron: derivative
			{
				// Act Function Derivative @ Y_y
				double derv_i = General_functions::DervActFcn(y_i(hid), actFcnFlatness, NeuronActFunct); //novo nome

				//dV/dyi = dij[hid] * Vij_out * (-1.0/(2.0*sig2*TrainDimIn*TrainDimIn))
				double error_i = dij(hid) * Vij_out * (-1.0/(2.0*sig2*dTrainDim*dTrainDim))*derv_i; //novo nome 

				//dyi/dw = derv_ActFn_t * x_t
				for (unsigned short in=0; in<DimIn; in++)
				{
					double momterm = momRate * ( Weights1(in,hid) - prev_Weights1(in,hid)); //momentum term
					delta_weights1(in,hid) = ( (-1.0) * lrCoeff1 * (error_i * TrainSet(i,in)) ) + momterm ; //MaxE = Min IP				
					Weights1(in,hid) += delta_weights1(in,hid);
				}

				Bias1(hid) += ( (-1.0) * lrCoeff1 * (error_i) ) + ( momRate * ( Bias1(hid) - prevBias1(hid) ) );
			}
		}
	}
	//##### End of BackForward Step == MaxE Criterium #####
	
}

//=================================
void CostFunction_Ent (
	double &CostFn,  //para ser atualizada diretamente
	const unsigned short &DimHidden, 
	const unsigned short &NumPatterns, 
	const mat &HiddenMatrix, 
	const double &sig2)
{
	double aux_CostFn = 0.0;
	
	double dNumPatterns = static_cast<double>(NumPatterns);

	rowvec dij3(DimHidden);
	for (unsigned short i=0; i<NumPatterns; i++) //usa-se indice i e j por percorrer as combinaes dois a dois
	{
		for (unsigned short j=0; j<NumPatterns; j++)
		{
			for (unsigned short hid=0; hid<DimHidden; hid++)
			{
				dij3(hid) = HiddenMatrix(i,hid) - HiddenMatrix(j,hid);
			}
			aux_CostFn += General_functions::Vij_function(dij3, DimHidden, sig2);
		}
	}
	
	CostFn = -log10 (( 1.0 / (dNumPatterns*dNumPatterns) ) * aux_CostFn);
	
}