//=================================
// include guard
#pragma once

//=================================
// included dependencies
#include <vector>
#include <armadillo>
#include <boost/random.hpp>

#include "..\enumerations.h"

//=================================
// namespaces
using namespace std;
using namespace arma;


class General_functions
{
public:

	//=================================
	//~General_functions(void); // destructor

	//=================================
	//General_functions(void); // constructor, default

	//#define W1_file
	//#define W2_file
	//#define B1_file
	//#define B2_file

	//=================================
	//mt importante, tem o MSE do train, test e output, finais (ao fim de todas as epocas)
	// o melhor para cada seed, de entre todos os c1 e c2 testados
	#ifndef MSE_Validation_file
	#define MSE_Validation_file
	#endif

	//=================================
	//resultados ao fim de todas as epocas
	#ifndef ResultsFile
	#define ResultsFile
	#endif

	//=================================
	//resultados para cada epoca (MSE)
	#ifndef DetailsFile
	#define DetailsFile
	#endif

	//=================================
	//retorna para cada epoca o valor da cost function da primeira parte (so p itl methods)
	#ifndef CostFnTestFile
	#define CostFnTestFile
	#endif

	//=================================
	//ficheiro com o tempo em segundos para todos os c1 e c2, so para o treino da rede
	#ifndef TimeFile
	#define TimeFile
	#endif

	//=================================
	//sai as duas matrizes de validation a pura e a 100%, para o estudo com denoising ou sem (puro), caso jakov
	//sai um ficheiro com o mse, a seed e a matriz de outputs dos target 
	#ifndef ValidationOutputFiles
	#define ValidationOutputFiles
	#endif

	//=================================
	//definio da funo de aleatorios entre -1 e 1
	//boost::random::uniform_real_distribution<> aleat(-1,1);

	//=================================
	//Helper function for MULTI kernel calculations
	static double gaussKernel(
		const double &e, 
		const double &sig2, 
		const unsigned short &m);


	//=================================
	static double Vij_function(
		const rowvec &dij, 
		const unsigned short &Dim, 
		const double &sig2); // entropy


	//=================================
	static double  DervActFcn(
		double output, //tem de ficar copia
		const float &actFcnFlatness, 
		const ActivationFunctions &NeuronActFunct);
	

	//=================================
	//functions of initialization of weights
	static void PCA_function(
		mat &Weights1, //para ser actualizada
		const mat &TrainSet_PCA, 
		const unsigned short &DimIn, 
		const unsigned short &DimHidden);

	//=================================
	//read files
	static mat lerFicheiro	(
		const unsigned short &linhas, 
		const unsigned short &colunas, 
		const string &Filename);


	//=================================
	static double  actFcn(const double &activation, 
		const float &actFcnFlatness, 
		const ActivationFunctions &NeuronActFunct);


	//=================================
	static void transitionLayer (
		mat &AfterMatrix, //para ser actualizado
		const unsigned short &NumberPatterns, 
		const unsigned short &DimFirstLayer, 
		const unsigned short &DimSecondLayer, 
		const mat &originalMatrix, 
		const mat &Weights, 
		const rowvec &Bias, 
		const float &actFcnFlatness, 
		const ActivationFunctions &NeuronActFunct);



	//=================================
	static mat nnOutputsFunction (
		const unsigned short &NumberPatterns, 
		const unsigned short &DimIn, 
		const unsigned short &DimHidden, 
		const unsigned short &DimOut, 
		const mat &Weights1, 
		const mat &Weights2, 
		const rowvec &Bias1, 
		const rowvec &Bias2, 
		const float &actFcnFlatness, 
		const mat &InMatrix, 
		const ActivationFunctions &NeuronActFunct1, 
		const ActivationFunctions &NeuronActFunct2);

	//=================================
	//error function
	static double MAEfunction (
		const unsigned short &linhas, 
		const unsigned short &colunas, 
		const mat &ErrorMatrix);

	//=================================
	static double MSEfunction (
		const unsigned short &linhas, 
		const unsigned short &colunas, 
		const mat &ErrorMatrix);


	//=================================
	static double sigmaFunction (
		const unsigned short &DimHidden, 
		const unsigned short &TrainDim, 
		const mat &HiddenTrain, 
		const sigmaOption &sigmaMethod,
		const float &AuxSigma);

	//=================================
	static float maxLrCoeffFunction (
		const unsigned short &DimIn, 
		const mat &TrainSet, 
		const double &learningInicial);


	//=================================
	static float SatLevelFunction (
		const unsigned short &TrainDim, 
		const unsigned short &colunas, 
		const mat &Matrix, 
		const float &SatLimit);

	//=================================
	static void Saturation_Process (
		const unsigned short &TrainDim, 
		const unsigned short &linhas, 
		const unsigned short &colunas, 
		const mat &TrainSet,
		const float &SatLimit,
		const float &ThresholdSat,
		const float &actFcnFlatness, 
		const ActivationFunctions &NeuronActFunct1,
		mat &Evaluated_Matrix, 
		mat &Weights1,
		rowvec &Bias1,
		boost::random::mt19937 &rng	);

	//=================================
	static mat ErrorMatrixFunction (
		const unsigned short &linhas, 
		const unsigned short &colunas, 
		const mat &Matrix1, 
		const mat &Matrix2);

	//=================================
	static double Error_MSE_function (
		const unsigned short &linhas, 
		const unsigned short &colunas, 
		const mat &Matrix1, 
		const mat &Matrix2);

	//=================================
	static string doubletoStringFunction (double a);

	//=================================
	static string StringConstructor_function (
		const double &c1, 
		const string &MethodName, 
		const string &tipo);

	//=================================
	static string StringTrainMethods (const TrainMethods_W1 &Method);

	//=================================
	static void learningAdaptCoeff	(
		const double &CostFn, 
		const double &Prev_CostFn, 
		float &lrCoeff, //esta variavel  actualizada directamente	
		const float &lr_inc, 
		const float &Max_lrCoeff, 
		const float &max_perf_inc, 
		const float &lr_dec, 
		const int &maxORmin	);

	//=================================
	static void learningAdaptMatrix(
		const double &CostFn, 
		const double &Prev_CostFn, 
		const float &max_perf_inc, 
		mat &Matrix, //vai ser alterada a propria matriz de pesos
		const mat &Prev_Matrix, 
		const int &maxORmin);

	//=================================
	static void learningAdaptBias(
		const double &CostFn, 
		const double &Prev_CostFn, 
		const float &max_perf_inc, 
		rowvec &Bias, //actualiza o BIAS
		const rowvec &Prev_Bias, 
		const int &maxORmin);

	//=================================
	//funo de estimar o cim para matrizes, no  usado mas pode ser util um dia
	static double CIM_Estimator (
		const unsigned short &DimOut, 
		const unsigned short &NumPatterns, 
		const mat &OutMatrixS,
		const mat &OutMatrixO, 
		const double &sig2);

	//=================================
	static double Corr_Estimator_fromVectors (
		const vec &vetor1, 
		const vec &vetor2, 
		const unsigned short &DimOut, 
		const double &sig2);

	//=================================
	static double CIM_Estimator_fromVectors (
		const unsigned short &DimOut, 
		const vec &OutvecS,
		const vec &OutvecO, 
		const double &sig2);



	//=================================
	//CASO W1B1: Dim2=DimHidden, Dim1=dimIn; CASO W2B2: Dim2=dimOut, Dim1=DimHidden
	static void OjaFunction (
		const unsigned short &Dim2, 
		const unsigned short &Dim1, 
		mat& Weights, 
		rowvec& Bias);
};
