//=================================
// include guard
#pragma once

//=================================
// included dependencies
#include <iostream>
#include <math.h>
#include <stdio.h>
#include <vector>
#include <stdlib.h>
#include <armadillo>
#include "General_Functions.h"
#include "..\enumerations.h"

//boost libraries
#include <boost/random.hpp>

#include <algorithm> //sort
#include <time.h> //time count

//=================================
// namespaces
using namespace std;
using namespace arma;

//=================================
void NetworkPROP_function(
	mat &W1, // para ser alterada directamente
	rowvec &B1, // para ser alterada directamente
	mat &W2, // para ser alterada directamente
	rowvec &B2, // para ser alterada directamente
	const InitializationW1Methods &Init_W1_Method, 
	const InitializationW2Methods &Init_W2_Method,
	const InitializationB1Methods &Init_B1_Method,
	const InitializationB2Methods &Init_B2_Method,
	const string &Label_data, 
	const ActivationFunctions &NeuronActFunct1,
	const ActivationFunctions &NeuronActFunct2,
	const OjaRule &OjaDecision,
	const unsigned short &seed, 
	boost::random::mt19937 &rng, //para alterar directamente 
	const unsigned short &DimIn, 
	const unsigned short &DimHidden, 
	const unsigned short &DimOut, 
	const unsigned short &TrainDim, 
	const unsigned short &TestDim, 
	const unsigned short &ValidationDim, 
	const vector <float> &learnCoeff1,
	const mat &TrainSet, 
	const mat &TestSet, 
	const mat &ValidationSet, 
	const mat &TargetTrSet,
	const mat &TargetTsSet,
	const mat &TargetVlSet,
	const unsigned short &numEpochs, 
	const float &lr_dec, 
	const float &lr_inc, 
	const float &max_perf_inc, 
	const float &goal,
	const float &momRate, 
	const float &actFcnFlatness, 
	const float &SatLimit, 
	const float &ThresholdSat)
{
	boost::random::uniform_real_distribution<> aleat(-1,1);

	#ifdef TimeFile
		time_t start,end; //count TIME by epoch
		time (&start); //TIME
	#endif

	//###defining a vector containning the learning coefficients to test the redutor###
		
	unsigned int DimLC;
	DimLC = learnCoeff1.size();
	//###end of defining a vector containning the learning coefficients to test the redutor###
	

	ofstream fileV;//ficheiros de saida de dados
	
	#ifdef ResultsFile
		string name = Label_data;
		name.append("Results_Prop_");
		name.append(".csv");
	#endif
		
	double dDimIn = static_cast<double>(DimIn); 
	double dDimHidden = static_cast<double>(DimHidden); 
	double dDimOut = static_cast<double>(DimOut); 
	double dTestDim = static_cast<double>(TestDim); 
	double dTrainDim = static_cast<double>(TrainDim); 
	
	////To have the same random initialization FOR each experiment without depending on choices of B1, B2, W1, W2
	rowvec Bias1Fixo(DimHidden);
	rowvec Bias2Fixo(DimOut);
	mat Weights1Fixo(DimIn, DimHidden);
	mat Weights2Fixo(DimHidden,DimOut);

	for (unsigned short hid=0; hid<DimHidden; hid++)
	{
		Bias1Fixo(hid) = aleat(rng);
	}
	for (unsigned short out=0; out<DimOut; out++)
	{
		Bias2Fixo(out) = aleat(rng);
	}
	for (unsigned short in=0; in<DimIn; in++)
	{
		for (unsigned short hid=0; hid<DimHidden; hid++)
		{
			Weights1Fixo(in,hid) = aleat(rng);
		}
	}
	for (unsigned short hid=0; hid<DimHidden; hid++)
	{
		for (unsigned short out=0; out<DimOut; out++) 
		{
			Weights2Fixo(hid,out) = aleat(rng);
		}
	}


	
	//##### Bias1 Inicialization #####
	switch (Init_B1_Method)
	{
		case RANDOM_B1: 
		{
			break;
		}
		
		case ZEROS_B1:
		{
			Bias1Fixo.zeros();
			break;
		}
			
		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}
	
	
	//Random Initialization of vector Bias2
	switch (Init_B2_Method)
	{
		case RANDOM_B2: 
		{
			break;
		}
		
		case ZEROS_B2:
		{
			Bias2Fixo.zeros();
			break;
		}
			
		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}

	
	//##### Weights1 Inicialization #####
	switch (Init_W1_Method)
	{
		case RANDOM_W1: 
		{
			break;
		}
		
		case PCA:
		{
			General_functions::PCA_function(Weights1Fixo, TrainSet, DimIn, DimHidden);
			break;
		}
			
		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}
	//##### End of Weights1 Inicialization #####
	

	//##### Weights2 Inicialization #####	
	//W2 may be initialized with Random or Transposed final Weights from 1st part
	switch (Init_W2_Method)
	{
		case RANDOM_W2: 
		{
			break;
		}
		
		case TRANSPOSE:
		{
			Weights2Fixo = trans(Weights1Fixo);
			break;
		}
			
		default:
		{
			cout << "ERROR!!! ";
			break;
		}
	}
	
	
	//as variveis improved so as escolhidas entre todas as combinao de c1 e c2
	mat ImprovedW1 (DimIn, DimHidden);
	mat ImprovedW2 (DimHidden, DimOut);
	rowvec ImprovedB1 (DimHidden);
	rowvec ImprovedB2 (DimOut);
	double Improved_MSEtest = 1000.0;

	for (unsigned int v=0; v<DimLC; v++)
	{
		//abrir ficheiros para incluso dos outputs detalhados aqui
		//assim abre um ficheiro para cada combinao c1 e c2, e coloca os resultados para todas as seeds desta combinao no mesmo ficheiro
		mat Weights1 =Weights1Fixo;
		rowvec Bias1=Bias1Fixo;	
		mat Weights2 =Weights2Fixo;
		rowvec Bias2=Bias2Fixo;	
			
		//##### OUTPUTS for each lrCoeff #####
		#ifdef DetailsFile
			ostringstream learn1;
			learn1 << learnCoeff1[v];
			string str1 = learn1.str();
			
			string detailed (Label_data);
			detailed.append("detailed_c1_");
			detailed.append(str1);
			detailed.append(".csv");
		#endif
			
		
		// #### Main Parameters Definition #####
		float lrCoeff1 = learnCoeff1[v];
		// #### End of Main Parameters Definition #####

		float Max_lrCoeff1 = General_functions::maxLrCoeffFunction (DimIn, TrainSet, learnCoeff1[v]);
		
		//##### Calculating HIDDEN and OUTPUT Matrices   #####
		//cria matriz de HiddenTrain de (Dimensao padres X tamanhoOutput)
		mat HiddenTrain (TrainDim, DimHidden);
		General_functions::transitionLayer (HiddenTrain, TrainDim, DimIn, DimHidden, TrainSet, Weights1Fixo, Bias1Fixo, actFcnFlatness, NeuronActFunct1);
		//##### End of Calculating Hidden Matrix #####

		
		mat prev_HiddenTrain (TrainDim,DimHidden);			
		prev_HiddenTrain = HiddenTrain;

		mat OutputsTrain (TrainDim, DimOut); 
		General_functions::transitionLayer (OutputsTrain, TrainDim, DimHidden, DimOut, HiddenTrain, Weights2Fixo, Bias2Fixo, actFcnFlatness, NeuronActFunct2);
			
		mat prev_OutputsTrain(TrainDim, DimOut);
		prev_OutputsTrain=OutputsTrain;
		//#####END of Calculating HIDDEN and OUTPUT Matrices  #####
				

		//##### Prop NN Training Inicialization #####
		mat delta_weights2(DimHidden,DimOut);
		mat delta_weights1(DimIn,DimHidden);		

	
		double MSE_train = 0.0;

		double Prev_CostFn = 0.0, CostFn = 100000.0;//grande para criterio adaptativo do lrCoeff nao cortar na 1a iteraao
		double Best_MSE_test=100.0, Best_MSE_train=100.0;

		mat prev_Weights1(DimIn,DimHidden), prev_Weights2(DimHidden, DimOut);
		prev_Weights1.zeros(), prev_Weights2.zeros();
		rowvec prevBias1(DimHidden); prevBias1.zeros();
		rowvec prevBias2(DimOut); prevBias2.zeros();

		//Best_xxx = best variaveis de entre as pocas; 
		//CorrespondenteMSEtrain = MSE treino correspondente para a poca seleccionada
		mat Best_Weights1(DimIn,DimHidden);
		mat Best_Weights2(DimHidden, DimOut);
		rowvec Best_Bias1(DimHidden);
		rowvec Best_Bias2(DimOut);
		float CorrespondenteMSEtrain = 0.0f;

		for (unsigned int Epochs1=0; Epochs1<numEpochs; Epochs1++)
		{
			//##### BackForward Step == Prop Criterium #####
			for (unsigned int tr=0; tr<TrainDim; tr++) //Cross samples: sample t
			{
				// #################  2a PARTE #################
				// dV/Dw = dV/dy * dy/dw
					
				for(unsigned int out=0; out<DimOut; out++) //for each output neuron: derivative
				{
					//dV/dy //ERRO QUADRATICO MEDIO COMO COST FUNNCTIOn
					double error = - ( 2.0/(dTrainDim*dDimOut) ) * (TargetTrSet(tr,out) - OutputsTrain(tr,out));

					//dy/dw = derv_ActFn_t * x_t
					for (unsigned int hid=0; hid<DimHidden; hid++)
					{
						double momterm = momRate * ( Weights2(hid,out) - prev_Weights2(hid,out)); //momentum term
						delta_weights2(hid,out) = (-1.0) * lrCoeff1 * (error * General_functions::DervActFcn( OutputsTrain(tr,out), actFcnFlatness, NeuronActFunct2) * HiddenTrain(tr,hid) ) + momterm;
							
						prev_Weights2(hid,out) = Weights2(hid,out);//momentum term
						Weights2(hid,out) += delta_weights2(hid,out);
					}
						
					prevBias2(out) = Bias2(out);
					Bias2(out) += ((-1.0) * lrCoeff1 * error * General_functions::DervActFcn(OutputsTrain(tr,out), actFcnFlatness, NeuronActFunct2)) + momRate * (Bias2(out) - prevBias2(out));
				}
				// #################  FIM da 2a PARTE #################
										

				// #################  1a PARTE #################
				// dE/Dw = dE/dy2 * dy2/dnet2 * dnet2/dy1 * dy1/dnet1 *dnet1/dw
				double aux4 = 0.0;
				vector<double> DervInner;
				DervInner.clear();

				for(unsigned int hid=0; hid<DimHidden; hid++) //for each output neuron: derivative
				{
					for(unsigned int out=0; out<DimOut; out++)
					{
						//dE/dy2
						double error = - ( 2.0/(dTrainDim*dDimOut) ) * (TargetTrSet(tr,out) - OutputsTrain(tr,out));
					
						//dy2/dnet2
						double D2 = General_functions::DervActFcn( OutputsTrain(tr,out), actFcnFlatness, NeuronActFunct2);

						//dnet2/dy1
						double D3 = Weights2(hid,out);
								
						aux4 += error * D2 * D3;
					}
					DervInner.push_back(aux4);
					aux4 = 0.0;
					
					//dy1/dnet1
					double D4 = General_functions::DervActFcn( HiddenTrain(tr,hid), actFcnFlatness, NeuronActFunct1);

					for(unsigned int in=0; in<DimIn; in++)
					{
						//dnet1/dw
						double D5 = TrainSet(tr,in);

						double momterm = momRate * ( Weights1(in,hid) - prev_Weights1(in,hid)); //momentum term
						delta_weights1(in,hid) =  (-1.0) * lrCoeff1 * ( DervInner[hid] * D4 * D5 ) + momterm ;
							
						prev_Weights1(in,hid) = Weights1(in,hid);
						Weights1(in,hid) += delta_weights1(in,hid);
					}
					prevBias1(hid) = Bias1(hid);
					Bias1(hid) += (-1.0) * lrCoeff1 * (DervInner[hid]*D4) +  momRate * (Bias1(hid) - prevBias1(hid));
				}

					
				// #################  FIM da 1a PARTE #################
			} //fim dos patterns
			//##### End of BackForward Step == Prop Criterium #####


			switch (OjaDecision)
			{
				case OjaYES: 
				{
					//####Aplicao da OJA's rule
					General_functions::OjaFunction (DimHidden, DimIn, Weights1, Bias1);
					
					//####Aplicao da OJA's rule for W2
					General_functions::OjaFunction (DimOut, DimHidden, Weights2, Bias2);
					break;
				}
				case OjaNO: 

				break;
				
				default: cout << "ERROR!!! ";
			}


			// ##### Update HIDDEN and OUTPUT matrices #####
			// ##### HiddenTrain Matrix #####
			prev_HiddenTrain = HiddenTrain;
			General_functions::transitionLayer (HiddenTrain, TrainDim, DimIn, DimHidden, TrainSet, Weights1, Bias1, actFcnFlatness, NeuronActFunct1);
			
			//####### SATURATION process #####
			General_functions::Saturation_Process (TrainDim, DimIn, DimHidden, TrainSet, SatLimit, ThresholdSat, actFcnFlatness, NeuronActFunct1, HiddenTrain, Weights1, Bias1, rng);
			
			//####### OutputsTrain Matrix #####
			prev_OutputsTrain = OutputsTrain;
			General_functions::transitionLayer (OutputsTrain, TrainDim, DimHidden, DimOut, HiddenTrain,Weights2, Bias2, actFcnFlatness, NeuronActFunct2);

			//####### SATURATION process #####
			General_functions::Saturation_Process (TrainDim, DimHidden, DimOut, HiddenTrain, SatLimit, ThresholdSat, actFcnFlatness, NeuronActFunct2, OutputsTrain, Weights2, Bias2, rng);	
					
			//#######################calculo da Funao Custo ############################
			mat ErrorMatrixTrain = General_functions::ErrorMatrixFunction(TrainDim, DimOut, TargetTrSet, OutputsTrain);
						
			//Funao Custo do data set
			Prev_CostFn = CostFn;
			CostFn = 0.0;
			for(unsigned int tr=0; tr<TrainDim; tr++)
			{
				for(unsigned int in=0; in<DimOut; in++)
				{
					CostFn += ErrorMatrixTrain(tr,in)*ErrorMatrixTrain(tr,in);
				}
			}
			CostFn = CostFn/(dTrainDim*dDimOut); // CostFn = MSE
			//#######################END calculo da Funao Custo ############################


			//##### Stop Criterium #####
			if ( abs(/*Prev_CostFn- */CostFn) < goal)
			{
				Epochs1 = numEpochs;
			}
			//##### End Stop Criterium  #####

			//##### Learning Coeff Adaptive #####

			

			if( CostFn/Prev_CostFn > max_perf_inc) 
			{
					lrCoeff1 =  lrCoeff1 * lr_dec; //corta o learning rate
					Weights1 = prev_Weights1; //no aceita pesos 
					Weights2 = prev_Weights2;
					Bias1= prevBias1;
					Bias2= prevBias2;
					OutputsTrain=prev_OutputsTrain;
					HiddenTrain=prev_HiddenTrain;
					if (lrCoeff1 < 0.00001)
					{
						lrCoeff1 = 0.00001;
					}
			}
			if( CostFn/Prev_CostFn < 1.0)
			{
				lrCoeff1 =  lrCoeff1 * lr_inc; //aceita pesos e aumenta learning rate
				if(lrCoeff1 > Max_lrCoeff1)
				{lrCoeff1 = Max_lrCoeff1;}

			}
			if((CostFn/Prev_CostFn >= 1.0) && (CostFn/Prev_CostFn <= max_perf_inc))
			{
				//aceita os pesos
				//no altera o learning rate
			}
			//##### End Learning Coeff Adaptative #####

			
			#ifdef DetailsFile
				fileV.open(detailed, ios::app);
				fileV << "Seed; "<< seed <<";Epoca ;" << Epochs1 ;
				fileV.close();
			#endif

			//#######################calculo do MSE_train ############################
			//MSE_train = CostFn;
			MSE_train=CostFn;
			//BEST
			if (MSE_train<Best_MSE_train)
			{
				Best_MSE_train=MSE_train;
			} 


		
			#ifdef DetailsFile
				fileV.open(detailed, ios::app);
				fileV << "; MSE_Train ;" << MSE_train << "; Best_MSE_Train ;" << Best_MSE_train;
				fileV.close();
			#endif
		

						
			//compresso
			mat HiddenTest (TestDim, DimHidden);
			General_functions::transitionLayer (HiddenTest, TestDim, DimIn, DimHidden, TestSet, Weights1, Bias1, actFcnFlatness, NeuronActFunct1);
	
			//descompresso
			mat OutputsTest (TestDim, DimOut);
			General_functions::transitionLayer (OutputsTest, TestDim, DimHidden, DimOut, HiddenTest, Weights2, Bias2, actFcnFlatness, NeuronActFunct2);


			//#######################calculo do MSE_TESTE ############################
			double MSE_test = General_functions::Error_MSE_function (TestDim, DimOut, TargetTsSet, OutputsTest);
					
			//BEST
			if (MSE_test < Best_MSE_test)
			{
				Best_MSE_test=MSE_test;
				Best_Weights1= Weights1;
				Best_Weights2= Weights2;
				Best_Bias1= Bias1;
				Best_Bias2= Bias2;
				CorrespondenteMSEtrain = static_cast<float>(MSE_train);
			} 
				
			#ifdef DetailsFile
				fileV.open(detailed, ios::app);
				fileV <<"; MSE_test ;"<< MSE_test << "; Best_MSE_test ;"<< Best_MSE_test <<";"<< endl;
				fileV.close();			
			#endif
				// END Ficheiros sada
			//#######################FIM calculo do MSE_test ############################
				
	
		}//fim epocas1
		//##### End of PROP NN Training #####

		//funcao rede (W1,W2,B1,B2) retorna MSE_ValidationSet
		mat ValidationOutput = General_functions::nnOutputsFunction (ValidationDim, DimIn, DimHidden, DimOut, Best_Weights1, Best_Weights2, Best_Bias1, Best_Bias2, actFcnFlatness, ValidationSet, NeuronActFunct1, NeuronActFunct2);
		double MSEValidation = General_functions::Error_MSE_function (ValidationDim, DimOut, TargetVlSet, ValidationOutput);
		
		//ofstream fileV;

		#ifdef MSE_Validation_file;
			string valdFile(Label_data);
			valdFile.append("Prop_ValidationMSE");
			valdFile.append(".csv");
			fileV.open(valdFile, ios::app);
			fileV << "Seed:;" << seed <<";C1:;"<< learnCoeff1[v] << ";MSE_TrainCorrespondente;" << CorrespondenteMSEtrain << ";Best_MSE_Test:;"<< Best_MSE_test << ";MSE validationSet:;" << MSEValidation << endl;
			fileV.close();
		#endif

		//BEST
		if (Best_MSE_test < Improved_MSEtest)
		{
			Improved_MSEtest = Best_MSE_test;
			ImprovedW2 = Best_Weights2;
			ImprovedB2 = Best_Bias2;
			ImprovedW1 = Best_Weights1;
			ImprovedB1 = Best_Bias1;
		}
		
		#ifdef ResultsFile
			fileV.open(name, ios::app); 
			fileV << "lrCoeff1 Inicial:;" << learnCoeff1[v] << ";lrCoeff1 Final:;" << lrCoeff1 << ";seed:;" << seed << ";Best_MSE_test;" << Best_MSE_test << ";" << endl;
			fileV.close();
		#endif

		//detailed_output.close();
	} //fim do ciclo v = vector lrCoeff
	
	
	#ifdef JeanValidationOutputFiles
		//funcao rede (W1,W2,B1,B2) retorna MSE_ValidationSet
		mat IValidationOutput = nnOutputsFunction (ValidationDim, DimIn, DimHidden, DimOut, ImprovedW1, ImprovedW2, ImprovedB1, ImprovedB2, actFcnFlatness, ValidationSet, NeuronActFunct1, NeuronActFunct2);
		double IMSEValidation = Error_MSE_function (ValidationDim, DimOut, TargetVlSet, IValidationOutput);
			
		string valdFile(Label_data);
		valdFile.append("Prop_ValidationMSE_and_Output");
		valdFile.append(".csv");
		fileV.open(valdFile, ios::app);
		fileV << "Seed:;" << seed << ";MSE_Validation;" << IMSEValidation << endl;
		for (unsigned int i=0; i<ValidationDim;i++)
		{
			for (unsigned int j=0; j < DimOut; j++)
			{
				fileV << IValidationOutput(i,j) << ";";
			}
			fileV << endl;
		}
		fileV << endl;
		fileV.close();
	#endif

	//actualiza no main as seguintes matrizes e vectores:
	W1=ImprovedW1;
	B1=ImprovedB1;
	W2=ImprovedW2;
	B2=ImprovedB2;


	#ifdef TimeFile //TIME
		time (&end);
		string time = Label_data;
		time.append ("_Time.csv");
		fileV.open(time, ios::app);
		fileV << "seed:;" << seed <<  ";#C1:;" <<DimLC <<  ";seconds;" <<  difftime (end,start) << ";DimIn;" << DimIn <<";DimHidden;" << DimHidden << ";" << endl; //TIME
		fileV.close();
	#endif
}