#include <math.h>
#include <stdio.h>

#include <vector>
#include <stdlib.h>

#include <armadillo>

#include "General_Functions.h"
#include "..\enumerations.h"

using namespace std;
using namespace arma;

void Uni_Weights1Bias1_function (
	const unsigned short &DimHidden, 
	const unsigned short &DimIn, 
	const unsigned short &TrainDim, 
	const mat &HiddenTrain, 
	const mat &TrainSet, 
	const mat &Vij_x, 
	const colvec &Vi_x,
	const float &actFcnFlatness, 
	const double &sig2, 
	const float &momRate, 
	const float &lrCoeff1, 
	mat &Weights1, //para ser alterado directamente
	const mat &prev_Weights1, 
	rowvec &Bias1, //para ser alterado directamente
	const rowvec &prevBias1, 
	const ActivationFunctions &NeuronActFunct)
{
	//double dTrainDim = static_cast<double>(TrainDim);
	float fTrainDim = static_cast<float>(TrainDim);
	fmat delta_weights(DimIn,DimHidden);
	delta_weights.zeros();

	for (unsigned short n=0; n < DimHidden; n++)
	{
		for (unsigned short m=n+1; m< DimHidden; m++)
		{
			fmat Vij_x(TrainDim,TrainDim), Vij_y(TrainDim,TrainDim);
			fvec Vj_x(TrainDim), Vj_y(TrainDim);
			float aux_vjx = 0.0f, aux_vjy = 0.0f;
			rowvec dij_y (1), dij_x (1);
			dij_y.zeros(); dij_x.zeros();

			float aux_V_x = 0.0f, aux_V_y = 0.0f, V_x, V_y;

			for (unsigned short i=0; i<TrainDim; i++)
			{
				for (unsigned short j=0; j<TrainDim; j++)
				{
					dij_x(0)= HiddenTrain(i,m) - HiddenTrain(j,m);
					dij_y(0) = HiddenTrain(i,n) - HiddenTrain(j,n);

					Vij_x(i,j) = General_functions::Vij_function(dij_x, 1, sig2);	//possibilidade de conflito double float								
					Vij_y(i,j) = General_functions::Vij_function(dij_y, 1, sig2);

					aux_vjx += Vij_x(i,j);
					aux_vjy += Vij_y(i,j);
				}
				Vj_x(i) = aux_vjx/fTrainDim;
				Vj_y(i) = aux_vjy/fTrainDim;

				aux_V_x += Vj_x(i);
				aux_V_y += Vj_y(i);

				aux_vjx = 0.0f;
				aux_vjy = 0.0f;
			}
			V_x = aux_V_x / fTrainDim;
			V_y = aux_V_y / fTrainDim;

			// ##### Fim Calculo Vij_saidas #####

			// ##### Calculo Vj #####
			float aux_vj = 0.0f, aux_vmx = 0.0f, aux_vmy = 0.0f;
			for (unsigned short i=0; i<TrainDim; i++)
			{
				for (unsigned short j=0; j<TrainDim; j++)
				{
					aux_vj += Vij_x(i,j) * Vij_y(i,j);
					aux_vmx += Vij_x(i,j);
					aux_vmy += Vij_y(i,j);
				}
			}
			// ##### Fim Calculo Vj #####

			//##### BackForward Step == MaxMI_Cauchy-Swartz Criterium #####
			float x_t = 0.0f, x_s = 0.0f, y_t = 0.0f, y_s = 0.0f;
			float Fi_x=0.0f, Fi_y=0.0f;
			float N1_y=0.0f, D1=0.0f, N2_y=0.0f, D2_x=0.0f, D2_y=0.0f, N3_y=0.0f, D3=0.0f;
			float N1_x=0.0f, N2_x=0.0f, N3_x=0.0f;
			D1 = aux_vj;
			D2_x = aux_vmx;
			D2_y = aux_vmy;
			dij_y(0) = 0.0, dij_x(0) = 0.0;

			for (unsigned short t=0; t<TrainDim; t++) //Cross samples: sample t
			{
				for (unsigned short s=0; s<TrainDim; s++) //Cross samples: sample s
				{
					x_t = HiddenTrain(t,m);//neuronio saida 1 patt t
					x_s = HiddenTrain(s,m);//neuronio saida 1 patt s
					y_t = HiddenTrain(t,n);//neuronio saida 2 patt t
					y_s = HiddenTrain(s,n);//neuronio saida 2 patt s

					dij_x(0) = x_t - x_s;
					dij_y(0) = y_t - y_s;

					// ############ dMI/dyi = Fi
					// ver Principe Tutorial pag.25 Eq.28
					N1_x += Vij_x(t,s) * Vij_y(t,s) * dij_x(0);
					N2_x += Vij_x(t,s) * dij_x(0);
					N3_x += Vj_x(s) * Vij_y(t,s) * dij_x(0);

					N1_y += Vij_x(t,s) * Vij_y(t,s) * dij_y(0);
					N2_y += Vij_y(t,s) * dij_y(0);
					N3_y += Vj_x(s) * Vij_y(t,s) * dij_y(0);
					D3 += Vj_x(s) * Vj_y(s);
				}
				Fi_x = -1.0f/( fTrainDim*fTrainDim * sig2 ) * ( N1_x/D1 + N2_x/D2_x - 2.0f*N3_x/D3 );
				Fi_y = -1.0f/( fTrainDim*fTrainDim * sig2 ) * ( N1_y/D1 + N2_y/D2_y - 2.0f*N3_y/D3 );

				// ############ dMI/Dw = dMI/dy * dy/dw ############				

					
				// Act Function Derivative @ Y_s
				//double derv_s = DervActFcn(y_s, actFcnFlatness);
				float dervx_t = General_functions::DervActFcn(static_cast<double>(x_t), actFcnFlatness, NeuronActFunct);
				float dervy_t = General_functions::DervActFcn(static_cast<double>(y_t), actFcnFlatness, NeuronActFunct);

				//dMI/dyi = Fi
				//dy/dw = derv_ActFn_t * x_t - derv_ActFn_s * x_s
				for (unsigned short i=0; i<DimIn; i++)
				{
					#if 0
					double momterm_x = momRate * ( Weights1(i,m) - prev_Weights1(i,m)); //momentum term
					double momterm_y = momRate * ( Weights1(i,n) - prev_Weights1(i,n)); //momentum term
					#endif

					delta_weights(i,m) = ( -1.0f * lrCoeff1 * Fi_x * (dervx_t * TrainSet(t,i) ) ) /*+ momterm_x*/ ; //Min MI entre neuronios = Min (-log(V(y)) = Max log V(y)
					delta_weights(i,n) = ( -1.0f * lrCoeff1 * Fi_y * (dervy_t * TrainSet(t,i) ) ) /*+ momterm_y */;

					//prev_Weights1(i,m) = Weights1(i,m); //REVER no ha necessidade de fazer esta actualizaao qd ela esta ja feita antes
					Weights1(i,m) += delta_weights(i,m);

					//prev_Weights1(i,n) = Weights1(i,n);
					Weights1(i,n) += delta_weights(i,n);
				}
				//prevBias1(m) = Bias1(m);
				//prevBias1(n) = Bias1(n);
					
				Bias1(m) += ( -1.0 * lrCoeff1 * Fi_x * dervx_t ) /*+ ( momRate * ( Bias1(m) - prevBias1(m) ) )*/;
				Bias1(n) += ( -1.0 * lrCoeff1 * Fi_y * dervy_t ) /*+ ( momRate * ( Bias1(n) - prevBias1(n) ) )*/;
			}
		}
	}
}



void CostFunction_Uni (
	double &Uni_CostFn, //para ser alterado diretamente
	const unsigned short &DimHidden, 
	const unsigned short &NumPatterns, 
	const mat &HiddenMatrix, 
	const double &sig2, 
	const mat &Vij_x, 
	const colvec &Vi_x, 
	const double &V_x)
{
	//double dNumPatterns = static_cast<double>(NumPatterns);
	float fNumPatterns=static_cast<float>(NumPatterns);
	float total_CostFn = 0.0;
	
	for (unsigned short n=0; n< DimHidden; n++)
	{
		for (unsigned short m=n+1; m< DimHidden; m++)
		{
			mat Vij_x(NumPatterns,NumPatterns), Vij_y(NumPatterns,NumPatterns);
			vec Vj_x(NumPatterns), Vj_y(NumPatterns);
			float aux_vjx = 0.0f, aux_vjy = 0.0f;
			rowvec dij_y (1), dij_x (1);
			dij_y.zeros(); dij_x.zeros();
			float aux_V_x = 0.0f, aux_V_y = 0.0f, V_x, V_y;

			for (unsigned short i=0; i<NumPatterns; i++)
			{
				for (unsigned short j=0; j<NumPatterns; j++)
				{
					dij_x(0) = HiddenMatrix(i,m) - HiddenMatrix(j,m);
					dij_y(0) = HiddenMatrix(i,n) - HiddenMatrix(j,n);

					Vij_x(i,j) = General_functions::Vij_function(dij_x, 1, sig2);									
					Vij_y(i,j) = General_functions::Vij_function(dij_y, 1, sig2);

					aux_vjx += Vij_x(i,j);
					aux_vjy += Vij_y(i,j);
				}
				Vj_x(i) = aux_vjx/fNumPatterns;
				Vj_y(i) = aux_vjy/fNumPatterns;

				aux_V_x += Vj_x(i);
				aux_V_y += Vj_y(i);

				aux_vjx = 0.0;
				aux_vjy = 0.0;
			}
			V_x = aux_V_x / fNumPatterns;
			V_y = aux_V_y / fNumPatterns;
			// ##### Fim Calculo Vij_saidas j com update #####
							
							
			//clculo do cost function
			float aux_VJ_CF = 0.0f, aux_VC_CF = 0.0f, CostFn = 0.0f;
			float VJ_CF=0.0f, VM_CF=0.0f, VC_CF=0.0f; //the 3 parts of the cost function: joint, marginal and cross information potentials
		 
			for (unsigned short i=0; i<NumPatterns; i++)
			{
				for (unsigned short j=0; j<NumPatterns; j++)
				{
					aux_VJ_CF += Vij_x(i,j) * Vij_y(i,j);
				}
				aux_VC_CF += Vj_x(i) * Vj_y(i);
			}

			VJ_CF = aux_VJ_CF / ( fNumPatterns*fNumPatterns ); //the argument for the joint information potential
			aux_VJ_CF = 0.0f;
			VM_CF = V_x * V_y; //the argument for the marginal information potential
			VC_CF = aux_VC_CF / ( fNumPatterns ); //the argunment for the cross information potential
			aux_VC_CF = 0.0f;

			CostFn = (log(VJ_CF) + log(VM_CF) - 2.0f*log(VC_CF)); //the complete information potential 
			//########## End of Calculating Cost Function for this Output ##########
			total_CostFn += CostFn;
		}
	}

	Uni_CostFn= (static_cast<double>(total_CostFn));
}