#pragma once

#include <iostream>
#include <math.h>
#include <stdio.h>

#include <vector>
#include <stdlib.h>

#include <armadillo>

#include "General_Functions.h"
#include "..\enumerations.h"

using namespace std;
using namespace arma;


//Sig_corr update
void Adaptive_Sigma_function (
	double &sig2_corr,//actual sig_corr para ser actualizado
	const vec &error_vector,//vec dos erros das L epocas anteriores + erro epoca actual
	const float &lrCoeff_sigma,
	const unsigned short &L, //window size == qts sig anteriores devem influenciar o actual
	const unsigned short &Epochs2
	)
{
	double sigma = sqrt(sig2_corr);
	double aux1=0.0, aux2=0.0;
	unsigned short L_aux = L;

	if (Epochs2==0)
	{
		cerr << "ERROR: you can not call this function for epochs=0" << endl;
		exit(1);
	}
	else if (Epochs2<L)
	{
		L_aux = Epochs2;
	}

	for(unsigned short i=L-L_aux ; i<L; i++)//percorre vec dos erros
	{
		aux1 += exp( ( -(error_vector(L)-error_vector(i))*(error_vector(L)-error_vector(i)) ) / ( 2*sigma*sigma ) )*
			( (error_vector(L)-error_vector(i))*(error_vector(L)-error_vector(i)) / ( sigma*sigma*sigma ) - 1/sigma );

		aux2 +=exp( ( -(error_vector(L)-error_vector(i))*(error_vector(L)-error_vector(i)) ) / ( 2*sigma*sigma ) );
	}

	sigma += lrCoeff_sigma * aux1/aux2;

	sig2_corr = sigma*sigma;
}


void W2_Corr_function (
	const unsigned short &TrainDim,
	const unsigned short &DimOut,
	const unsigned short &DimHidden,
	const mat &TargetTrSet,
	const mat &OutputsTrain,
	const mat &HiddenTrain,
	const double &sig2,
	const float &lrCoeff2, 
	const float &actFcnFlatness, 
	const ActivationFunctions &NeuronActFunct2,
	const OjaRule &OjaDecision,
	mat &Weights2,
	rowvec &Bias2
	)
{
	//estamos a MAX corr...(-1.0) sai da equaao dos delta_weights2 e Bias2
	vector <double> dervPI2;
	dervPI2.resize(DimOut);
	mat delta_weights2(DimHidden,DimOut); 

	for (unsigned short tr=0; tr<TrainDim; tr++)
	{
		//calcula a correntropia deste pattern::
		for(unsigned short out=0; out<DimOut; out++)
		{
			//derivada do CIM em ordem a yi
			dervPI2[out] = (1.0/static_cast<double>(DimOut))*
				(1.0/(sqrt(2.0*3.14159265359)*pow(sig2,3.0/2.0)))*
				(TargetTrSet(tr,out)-OutputsTrain(tr,out))*
				exp(-(TargetTrSet(tr,out)-OutputsTrain(tr,out))*(TargetTrSet(tr,out)-OutputsTrain(tr,out))/ (2*sig2));
		}
						
		//nesta altura est calculada a derivada de PI em ordem a yi
		//a seguir faz a derivada do yi em anlise em ordem aos pesos e multiplica pela dPY/dyi e por rho, com o sinal de acordo com maximizar ou minimizar

		for (unsigned short hid=0; hid<DimHidden; hid++)
		{
			for (unsigned short out=0; out<DimOut; out++)
			{
				//diferena das derivadas de y em ordem as pesos
				delta_weights2(hid,out) = /*(-1.0)**/ lrCoeff2 * dervPI2[out] * ( HiddenTrain(tr,hid) * General_functions::DervActFcn(OutputsTrain(tr,out), actFcnFlatness, NeuronActFunct2) );	
				Weights2(hid,out) += delta_weights2(hid,out);
			}
		}
					
		//nota: a saida do espao reduzido do neuronio bias  sempre 1
		for (unsigned short in=0; in<DimOut; in++)
		{
			Bias2(in) += /*(-1.0)**/ lrCoeff2 * (dervPI2[in]) * General_functions::DervActFcn(OutputsTrain(tr,in), actFcnFlatness, NeuronActFunct2);
		}
							

		switch (OjaDecision)
		{
			case OjaYES: 
			{
				//####Aplicao da OJA's rule for W2
				General_functions::OjaFunction (DimOut, DimHidden, Weights2, Bias2);
				break;
			}
								
			case OjaNO: 
				break;
				
			default: cout << "ERROR!!! ";
		}

		//fim da actualizao dos pesos para o vector yi
		//dervPI2.resize(DimIn);
	}
}


void W2_CIM_function(
	const unsigned short &TrainDim,
	const unsigned short &DimOut,
	const unsigned short &DimHidden,
	const mat &TargetTrSet,
	const mat &OutputsTrain,
	const mat &HiddenTrain,
	const double &sig2,
	const float &lrCoeff2,
	const float &actFcnFlatness, 
	const ActivationFunctions &NeuronActFunct2,
	const OjaRule &OjaDecision, 
	mat &Weights2, 
	rowvec &Bias2)
{
	vector <double> dervPI2;
	dervPI2.resize(DimOut);
	mat delta_weights2(DimHidden,DimOut); 

	for (unsigned short tr=0; tr<TrainDim; tr++)
	{
		//calcula a correntropia deste pattern::
		vec targetTrtSetV = TargetTrSet(tr,span());  //new
		vec trainOutV = OutputsTrain(tr,span());
		double CIM_xy = General_functions::CIM_Estimator_fromVectors (DimOut, targetTrtSetV, trainOutV, sig2);
									
		for(unsigned short out=0; out<DimOut; out++)
		{
			//derivada do CIM em ordem a yi
			dervPI2[out] = -0.5*(1.0/CIM_xy)*
				(1.0/static_cast<double>(DimOut))*
				(1.0/(sqrt(2.0*3.14159265359)*pow(sig2,3.0/2.0)))*
				(TargetTrSet(tr,out)-OutputsTrain(tr,out))*
				exp(-(TargetTrSet(tr,out)-OutputsTrain(tr,out))*(TargetTrSet(tr,out)-OutputsTrain(tr,out))/ (2*sig2))
				;
		}
						
		//nesta altura est calculada a derivada de PI em ordem a yi
		//a seguir faz a derivada do yi em anlise em ordem aos pesos e multiplica pela dPY/dyi e por rho, com o sinal de acordo com maximizar ou minimizar

		for (unsigned short hid=0; hid<DimHidden; hid++)
		{
			for (unsigned short out=0; out<DimOut; out++)
			{
				//diferena das derivadas de y em ordem as pesos
				//
				delta_weights2(hid,out) = (-1.0)* lrCoeff2 * dervPI2[out] * ( HiddenTrain(tr,hid) * General_functions::DervActFcn(OutputsTrain(tr,out), actFcnFlatness, NeuronActFunct2) );	
				Weights2(hid,out) += delta_weights2(hid,out);
			}
		}
					
		//nota: a saida do espao reduzido do neuronio bias  sempre 1
		for (unsigned short out=0; out<DimOut; out++)
		{
			Bias2(out) += (-1.0)* lrCoeff2 * (dervPI2[out]) * General_functions::DervActFcn(OutputsTrain(tr,out), actFcnFlatness, NeuronActFunct2);
		}
							

		switch (OjaDecision)
		{
			case OjaYES: 
			{
				General_functions::OjaFunction (DimOut, DimHidden, Weights2, Bias2);
				break;
			}
								
			case OjaNO: 
				break;
				
			default: cout << "ERROR!!! ";
		}

		//fim da actualizao dos pesos para o vector yi
		//dervPI2.resize(DimOut);
	}
}


void W2_Prop_function (
	const unsigned short &TrainDim, 
	const unsigned short &DimOut,
	const unsigned short &DimHidden,
	const mat &TargetTrSet,
	const mat &OutputsTrain,
	const mat &HiddenTrain,
	const float &lrCoeff2,
	const float &actFcnFlatness, 
	const ActivationFunctions &NeuronActFunct2,
	const OjaRule &OjaDecision, 
	mat &Weights2, 
	rowvec &Bias2 
	
	)
{
	vector <double> dervPI2;
	dervPI2.resize(DimOut);
	mat delta_weights2(DimHidden,DimOut); 
	//###########################dPI/dzk #####################################
	//derivada do potencial de informao em ordem a yi
	for (unsigned short tr=0; tr<TrainDim; tr++)
	{
		for(unsigned short out=0; out<DimOut; out++)
		{
			//1/2 do ERRO QUADRATICO (diferente de erro quadratico medio, ver depois)
			dervPI2[out] = -(TargetTrSet(tr,out) - OutputsTrain(tr,out));//new
		}
						
		//nesta altura est calculada a derivada de PI em ordem a yi
		//a seguir faz a derivada do yi em anlise em ordem aos pesos e multiplica pela dPY/dyi e por rho, com o sinal de acordo com maximizar ou minimizar

		for (unsigned short hid=0; hid<DimHidden; hid++)
		{
			for (unsigned short out=0; out<DimOut; out++)
			{
				//diferena das derivadas de y em ordem as pesos
				delta_weights2(hid,out) = (-1.0)* lrCoeff2 * dervPI2[out] * ( HiddenTrain(tr,hid) * General_functions::DervActFcn(OutputsTrain(tr,out), actFcnFlatness, NeuronActFunct2) );	
				Weights2(hid,out) += delta_weights2(hid,out);
			}
		}
					
		//nota: a saida do espao reduzido do neuronio bias  sempre 1
		for (unsigned short out=0; out<DimOut; out++)
		{
			Bias2(out) += (-1.0)* lrCoeff2 * (dervPI2[out]) * General_functions::DervActFcn(OutputsTrain(tr,out), actFcnFlatness, NeuronActFunct2);
		}

						
		switch (OjaDecision)
		{
			case OjaYES: 
			{
				//####Aplicao da OJA's rule for W2
				General_functions::OjaFunction (DimOut, DimHidden, Weights2, Bias2);
				break;
			}
								
			case OjaNO:
				break;
				
			default: cout << "ERROR!!! ";
		}	

		//fim da actualizao dos pesos para o vector yi
		dervPI2.resize(DimOut);
	}
}