
//boost libraries
#include <boost/random.hpp>

//armadillo
#include <armadillo>

//libraries
#include <vector>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <list>
#include <cstdlib>
#include <algorithm>
#include <ctime>

#include "hw/swarmHW.h"
#include "hw/myfitnessHW.h"

#include "EPSO\swarm.h"
#include "EPSO\myfitnessHWS.h"

#include "ITL_networks\NetworkFunction.h"
#include "ITL_networks\Normalization_functions.h"
#include "enumerations.h"
#include "ITL_networks\Prop_Network.h"


using namespace std;
using namespace arma;


void main()
{
	//#######################################################################################
	// PARAMETERS TO TRAIN THE NETWORK
	//#######################################################################################
	//defining the vectors containning the learning coefficients to test the redutor
	vector <float> learnCoeff1;
	vector <float> learnCoeff2;
	//=================================
	unsigned short numEpochs1;
	unsigned short numEpochs2;
	//=================================
	enum GLOBAL_Train Train_Global = PROP; //can vary between ITL or PROP 
	enum TrainMethods_W1 W1_Method; //ENTROPY, CAUCHY_SCHWARTZ, EUCLIDEAN_DISTANCE, UNI (definir o mtodo entre)
	enum TrainMethods_W2 W2_Method; //W2_PROP,	CIM,	Corr (definir o mtodo entre)
	//---------------------------------------------------------------
	const enum InitializationW1Methods Init_W1_Method= PCA; //definir o mtodo de inicializaao de W1 entre: RANDOM_W1, PCA
	const enum InitializationW2Methods Init_W2_Method = TRANSPOSE; //definir o mtodo de inicializaao de W2 entre: RANDOM_W2, TRANSPOSE
	const enum InitializationB1Methods Init_B1_Method = ZEROS_B1; // RANDOM_B1 or ZEROS_B1
	const enum InitializationB2Methods Init_B2_Method = ZEROS_B2; // RANDOM_B2 or ZEROS_B2
	const enum ActivationFunctions NeuronActFunct1 = NEURON_ACT_TANH; //funao de activaao para 1a camada:: NEURON_ACT_LOGISTIC, NEURON_ACT_TANH, NEURON_ACT_LINEAR
	const enum ActivationFunctions NeuronActFunct2 = NEURON_ACT_LINEAR; //funao de activaao para 2a camada:: NEURON_ACT_LOGISTIC, NEURON_ACT_TANH, NEURON_ACT_LINEAR
	const enum NormalizationMethod NormMethod = ENTRANCE_MIN_MAX; // defining normalization method: ENTRANCE_MIN_MAX, GLOBAL_MIN_MAX, ENTRANCE_ZSCORE, GLOBAL_ZSCORE, NONE
	const enum sigmaOption sigmaMethod = SILVERMAN_RULE; //	SILVERMAN_RULE or PRINCIPE_RULE
	const enum OjaRule OjaDecision = OjaNO; //can change betwee OjaYES and OjaNO
	//---------------------------------------------------------------
	unsigned int NumberSeeds=1;
	//---------------------------------------------------------------
	const float AuxSigma = 1.0f;//percentage of sigma from SilvermanRule
	//=================================
	string Label_data= ("PROP_");

	const float lr_dec = 0.5f;
	const float lr_inc = 1.05f;
	const float max_perf_inc = 1.04f; //performance - racio (=CostFunction/Prev_CostFunction) a partir do qual nao se actualizam pesos
	const float goal = 0.00000001f; //como no matlab
	const float momRate = 0.0f; //momentum coef
	const float actFcnFlatness = 1.0f; // For Activation Function

	const float SatLimit = 0.9f; //o valor da funo de activao a partir do qual se considera saturado
	const float ThresholdSat = 0.3f; //a percentagem mxima permitida de valores saturados a partir da qual se faz uma perturbao aleatria aos pesos
	//=================================
	//for sigma_corr
	const float AuxSigma2 = 1.0f;//for sigma_corr
	const unsigned short L = 4;//for sigma_corr
	float lrCoeff_sigma = 0.5f;//for sigma_corr 
	//=================================
	switch (Train_Global)
	{
		case PROP: 
		{
			numEpochs1 = 20;
			
			learnCoeff1.push_back(0.5f);
			//learnCoeff1.push_back(0.05f);
			//learnCoeff1.push_back(0.005f);
			//learnCoeff1.push_back(0.0005f);
			break;
		}
		case ITL: 
		{
			numEpochs1 = 2000;
			numEpochs2 = 2000;

			W1_Method = CAUCHY_SCHWARTZ; //ENTROPY, CAUCHY_SCHWARTZ, EUCLIDEAN_DISTANCE, UNI (definir o mtodo entre)
			W2_Method = W2_PROP; // W2_PROP, 	CIM, 	Corr

			//learnCoeff1.push_back(0.005f);
			learnCoeff1.push_back(0.5f);
			//learnCoeff1.push_back(0.00005f);
			/*learnCoeff1.push_back(0.0005f);
			learnCoeff1.push_back(0.00005f);*/

			//learnCoeff2.push_back(0.5f);
			//learnCoeff2.push_back(0.05f);
			learnCoeff2.push_back(0.005f);
			learnCoeff2.push_back(0.0005f);
			learnCoeff2.push_back(0.00005f);
			break;
		}
		case PCA_train: 
		{
			//nothing
			break;
		}
		default: cout << "ERROR!!! ";
	}
	//=================================
	// #### END Main Parameters Definition #####
	//=================================
	

	//#######################################################################################
	// PARAMETERS FOR HYBRID PROBLEM
	//#######################################################################################
	
	//=================================
	//			General Parameters
	//=================================
	enum ProblemSelection ProblemSelect = Reservoirs8; //Reservoirs8, Reservoirs12
	//=================================
	enum Initi_Vel	Vel_init = RANDOM_VEL; // 	TRANSFER_VEL 	 RANDOM_VEL
	
	vector<unsigned int> NumOfNeurons; // Vector holding the network structure
	
	unsigned int numParticles=100;
	//---------------------------------------------------------------
	unsigned int maxIterI=40; /*Nota: o maxIter  assim pequeno pq o treino s tem 300particulas e dp quero o swarm da iter 315 para continuar!!!!*/
	unsigned int maxIterS=40; //n de iteraes no espao reduzido
	unsigned int maxIterF=20;  //n de iteraes do enxame final!!!!
	unsigned int bigMaxIterI=maxIterI+maxIterS+maxIterF;//iterao at  qual quero ke o EPSO v no espao grande!!!!! (para comparar)
	//---------------------------------------------------------------
	unsigned int trainDim=10;/*NOTA IMPORTANTE: a alterao do terinoDim PODE IMPLICAR uma adaptao das var treinoBegin e treinoEnd*/
	unsigned int testDim=50;
	unsigned int validationDim=50;
	unsigned int ConjuntoDim=trainDim+testDim+validationDim;
	//---------------------------------------------------------------
	unsigned int TBegin=0, TEnd=maxIterI;
	//---------------------------------------------------------------
	bool minimize=false; // minimizing or maximizing
	//---------------------------------------------------------------
	#define TT_SetsSelection_EPSO //particulas seleccionadas pelas particulas do epso, ifndef ento sao aleatorias
	#define Hybrid_EPSO // quando definido faz o hybrid, quando indefinido faz o EPSO apenas no espao grande
	//=================================
	//			CASE STUDY:
	//=================================
	float numR, temp;
	unsigned int DimIn, DimHidden;//dimenso da camada interior
	unsigned short DimOut;
	float kturb, kpump, PHg1, PHg0, PGp1, PGp0, PHpw1, PHpw0, PWp1, PWp0, PWs1, PWs0;
	//---------------------------------------------------------------
	vec typeperiodEntry; vec windgenEntry;
	vec xprefinalEntry; vec xfinalEntry;
	fvec variaveisProblema;

	//xfinal=volume of water that exists inside of the reservoir in the end of the period

	//--------------------------------------------------------------
	// define o vetor capmax, numres e inflow;
	vec inflow,	capmin, capmax, qmin, qmax, numres;
	//X=pump/released water,
	//capmin=minimum capacity of the reservoir,
	//capmax=maximum capacity of the reservoir
	//inflow=rain water during period, xanterior=volume of water that exists inside of the reservoir in the begining of the period,
	//qmin= maximum amount of water that the turbine can pump
	//qmax= maximum amount of water that the turbine can release
	//---------------------------------------------------
	switch (ProblemSelect)
	{
		case Reservoirs8: 
		{
			numR=8.0f, temp=12.0f;
			DimIn=temp*numR;
			DimHidden=50;//dimenso da camada interior
			DimOut = DimIn;
			//---------------------------------------------------------------
			kturb=9.32f, kpump=12.25f;
			PHg1=0.08f, PHg0=0.01f, PGp1=0.10f, PGp0=0.02f, PHpw1=0.07f, PHpw0=0.005f, PWp1=0.07f, PWp0= 0.005f, PWs1=0.07f, PWs0=0.005f;
			//---------------------------------------------------------------
			typeperiodEntry.resize(DimIn);
			windgenEntry.resize(DimIn); windgenEntry.zeros();
			//typeperiod vector
			for (unsigned int i=1;i<(temp+1);i++)
			{
				if (i%2 != 0) { typeperiodEntry(span( (i-1)*numR, i*numR-1)).ones(); }  //fill numR positions with value 1
				else {	typeperiodEntry(span( (i-1)*numR, i*numR-1)  ).zeros(); }    //fill numR positions with value 0
			}	
			// Nota: Supomos que a wingen est em Kwh
			windgenEntry(0*numR)=6667.0; 
			windgenEntry(1*numR)=5333.0; 
			windgenEntry(2*numR)=10000.0; 
			windgenEntry(3*numR)=13333.0; 
			windgenEntry(4*numR)=11333.0;
			windgenEntry(5*numR)=6667.0; 
			windgenEntry(6*numR)=6000.0; 
			windgenEntry(7*numR)=8000.0; 
			windgenEntry(8*numR)=13333.0; 
			windgenEntry(9*numR)=8000.0;
			windgenEntry(10*numR)=6667.0; 
			windgenEntry(11*numR)=10000.0;
	
			xprefinalEntry.resize(DimIn+numR); xprefinalEntry.zeros();
			xfinalEntry.resize(DimIn+numR); xfinalEntry.zeros();
			//valor das primeiras 8 posioes do vector (ver excel), as restantes so definidas posteriormente
			xprefinalEntry[0]=10000; xfinalEntry[0]=10000;
			xprefinalEntry[1]=10000; xfinalEntry[1]=10000;
			xprefinalEntry[2]=10000; xfinalEntry[2]=10000;
			xprefinalEntry[3]=10000; xfinalEntry[3]=10000;
			xprefinalEntry[4]=10000; xfinalEntry[4]=10000;
			xprefinalEntry[5]=35000; xfinalEntry[5]=35000; 
			xprefinalEntry[6]=10000; xfinalEntry[6]=10000; 
			xprefinalEntry[7]=30000; xfinalEntry[7]=30000;
	
			variaveisProblema.resize(15);
			variaveisProblema (0) = kturb;
			variaveisProblema (1) = kpump;
			variaveisProblema (2) = PHg1;
			variaveisProblema (3) = PHg0;
			variaveisProblema (4) = PGp1;
			variaveisProblema (5) = PGp0;
			variaveisProblema (6) = PHpw1;
			variaveisProblema (7) = PHpw0;
			variaveisProblema (8) = PWp1;
			variaveisProblema (9) = PWp0;
			variaveisProblema (10) = PWs1;
			variaveisProblema (11) = PWs0;
			variaveisProblema (12) = numR;
			variaveisProblema (13) = temp;
			variaveisProblema (14) = DimHidden;
			//---------------------------------------------------
			// define o vetor capmax, numres e inflow;
			inflow.resize(DimIn), capmin.resize(DimIn), capmax.resize(DimIn), qmin.resize(DimIn), qmax.resize(DimIn), numres.resize(DimIn);

			for (unsigned int i=0; i<DimIn; i++) {capmin[i]=1000;}
	
			for (unsigned int i=0; i<DimIn; i+=8) {capmax[i]=15000; numres[i]=1; inflow[i]=133; qmin[i]=-20736.0; qmax[i]=24624.0;} //R1
			for (unsigned int i=1; i<DimIn; i+=8) {capmax[i]=20000; numres[i]=2; inflow[i]=0; qmin[i]=-20736.0; qmax[i]=24624.0;} //R2
			for (unsigned int i=2; i<DimIn; i+=8) {capmax[i]=25000; numres[i]=3; inflow[i]=0; qmin[i]=-27648.0; qmax[i]=32832.0;} //R3
			for (unsigned int i=3; i<DimIn; i+=8) {capmax[i]=15000; numres[i]=4; inflow[i]=133; qmin[i]=-20736.0; qmax[i]=24624.0;} //R4
			for (unsigned int i=4; i<DimIn; i+=8) {capmax[i]=20000; numres[i]=5; inflow[i]=0; qmin[i]=-20736.0; qmax[i]=24624.0;}
			for (unsigned int i=5; i<DimIn; i+=8) {capmax[i]=40000; numres[i]=6; inflow[i]=0; qmin[i]=-27648.0; qmax[i]=32832.0;}
			for (unsigned int i=6; i<DimIn; i+=8) {capmax[i]=15000; numres[i]=7; inflow[i]=133; qmin[i]=-20736.0; qmax[i]=24624.0;}
			for (unsigned int i=7; i<DimIn; i+=8) {capmax[i]=60000; numres[i]=8; inflow[i]=133; qmin[i]=0.0; qmax[i]=32832.0;}
			break;
		}
		case Reservoirs12: 
		{
			numR=12.0f, temp=12.0f;
			DimIn=temp*numR;
			DimHidden=50;//dimenso da camada interior
			DimOut = DimIn;
			//---------------------------------------------------------------
			kturb=9.32f, kpump=12.25f;
			PHg1=0.08f, PHg0=0.01f, PGp1=0.10f, PGp0=0.02f, PHpw1=0.07f, PHpw0=0.005f, PWp1=0.07f, PWp0= 0.005f, PWs1=0.07f, PWs0=0.005f;
			//---------------------------------------------------------------
			typeperiodEntry.resize(DimIn);
			windgenEntry.resize(DimIn); windgenEntry.zeros();
			//typeperiod vector
			for (unsigned int i=1;i<(temp+1);i++)
			{
				if (i%2 != 0) { typeperiodEntry(span( (i-1)*numR, i*numR-1)).ones(); }  //fill numR positions with value 1
				else {	typeperiodEntry(span( (i-1)*numR, i*numR-1)  ).zeros(); }    //fill numR positions with value 0
			}
			// Nota: Supomos que a wingen est em Kwh
			windgenEntry(0*numR)=6667.0; 
			windgenEntry(1*numR)=5333.0; 
			windgenEntry(2*numR)=10000.0; 
			windgenEntry(3*numR)=13333.0; 
			windgenEntry(4*numR)=11333.0;
			windgenEntry(5*numR)=6667.0; 
			windgenEntry(6*numR)=6000.0; 
			windgenEntry(7*numR)=8000.0; 
			windgenEntry(8*numR)=13333.0; 
			windgenEntry(9*numR)=8000.0;
			windgenEntry(10*numR)=6667.0; 
			windgenEntry(11*numR)=10000.0;
	
			xprefinalEntry.resize(DimIn+numR); xprefinalEntry.zeros();
			xfinalEntry.resize(DimIn+numR); xfinalEntry.zeros();
			//valor das primeiras 8 posioes do vector (ver excel), as restantes so definidas posteriormente
			xprefinalEntry[0]=10000; xfinalEntry[0]=10000;
			xprefinalEntry[1]=10000; xfinalEntry[1]=10000;
			xprefinalEntry[2]=10000; xfinalEntry[2]=10000;
			xprefinalEntry[3]=10000; xfinalEntry[3]=10000;
			xprefinalEntry[4]=10000; xfinalEntry[4]=10000;
			xprefinalEntry[5]=35000; xfinalEntry[5]=35000; 
			xprefinalEntry[6]=10000; xfinalEntry[6]=10000; 
			xprefinalEntry[7]=10000; xfinalEntry[7]=10000;
			xprefinalEntry[8]=30000; xfinalEntry[8]=30000;
			xprefinalEntry[9]=10000; xfinalEntry[9]=10000;
			xprefinalEntry[10]=10000; xfinalEntry[10]=10000;
			xprefinalEntry[11]=40000; xfinalEntry[11]=40000;
	
			variaveisProblema.resize(15);
			variaveisProblema (0) = kturb;
			variaveisProblema (1) = kpump;
			variaveisProblema (2) = PHg1;
			variaveisProblema (3) = PHg0;
			variaveisProblema (4) = PGp1;
			variaveisProblema (5) = PGp0;
			variaveisProblema (6) = PHpw1;
			variaveisProblema (7) = PHpw0;
			variaveisProblema (8) = PWp1;
			variaveisProblema (9) = PWp0;
			variaveisProblema (10) = PWs1;
			variaveisProblema (11) = PWs0;
			variaveisProblema (12) = numR;
			variaveisProblema (13) = temp;
			variaveisProblema (14) = DimHidden;

			//---------------------------------------------------
			// define o vetor capmax, numres e inflow;
			inflow.resize(DimIn), capmin.resize(DimIn), capmax.resize(DimIn), qmin.resize(DimIn), qmax.resize(DimIn), numres.resize(DimIn);

			for (unsigned int i=0; i<DimIn; i++) {capmin[i]=1000;}
	
			for (unsigned int i=0; i<DimIn; i+=8) {capmax[i]=15000; numres[i]=1; inflow[i]=133; qmin[i]=-20736.0; qmax[i]=24624.0;} //R1
			for (unsigned int i=1; i<DimIn; i+=8) {capmax[i]=20000; numres[i]=2; inflow[i]=0; qmin[i]=-20736.0; qmax[i]=24624.0;} //R2
			for (unsigned int i=2; i<DimIn; i+=8) {capmax[i]=25000; numres[i]=3; inflow[i]=0; qmin[i]=-27648.0; qmax[i]=32832.0;} //R3
			for (unsigned int i=3; i<DimIn; i+=8) {capmax[i]=15000; numres[i]=4; inflow[i]=133; qmin[i]=-20736.0; qmax[i]=24624.0;} //R4
			for (unsigned int i=4; i<DimIn; i+=8) {capmax[i]=20000; numres[i]=5; inflow[i]=0; qmin[i]=-20736.0; qmax[i]=24624.0;}
			for (unsigned int i=5; i<DimIn; i+=8) {capmax[i]=40000; numres[i]=6; inflow[i]=0; qmin[i]=-27648.0; qmax[i]=32832.0;}
			for (unsigned int i=6; i<DimIn; i+=8) {capmax[i]=15000; numres[i]=7; inflow[i]=133; qmin[i]=-20736.0; qmax[i]=24624.0;}
			for (unsigned int i=7; i<DimIn; i+=8) {capmax[i]=20000; numres[i]=8; inflow[i]=0; qmin[i]=-20736.0; qmax[i]=24624.0;}
			for (unsigned int i=8; i<DimIn; i+=8) {capmax[i]=60000; numres[i]=9; inflow[i]=133; qmin[i]=-27648.0; qmax[i]=32832.0;}
			for (unsigned int i=9; i<DimIn; i+=8) {capmax[i]=15000; numres[i]=10; inflow[i]=133; qmin[i]=-20736.0; qmax[i]=24624.0;}
			for (unsigned int i=10; i<DimIn; i+=8) {capmax[i]=20000; numres[i]=11; inflow[i]=0; qmin[i]=-20736.0; qmax[i]=24624.0;}
			for (unsigned int i=11; i<DimIn; i+=8) {capmax[i]=80000; numres[i]=12; inflow[i]=133; qmin[i]=0.0; qmax[i]=32832.0;}
			break;
		}
		default: cout << "ERROR!!! ";
	}
	//=================================
	//=================================

	//#### Random functions definition
	boost::random::mt19937 rng;
	//#### End of random functions definition ####
			
	HydroWind myFf;//funo fitness do espao grande
	#ifdef Hybrid_EPSO
		HydroWindHWS myFfs; //funo fitness do espao pequeno
	#endif
		
	//======================================
	//Defining seeds
	unsigned int GlobalSeedInitial=1245;
	unsigned int GlobalSeedPasso=50;

	vector <int> vecGlobalSeed;
	for (unsigned int gg=GlobalSeedInitial;gg<GlobalSeedInitial+GlobalSeedPasso*NumberSeeds; gg+=GlobalSeedPasso )
	{
		vecGlobalSeed.push_back(gg);
	}
	unsigned int SeedDim=vecGlobalSeed.size();
	

	//======================================
	//for space S
	vector <double> vec_tauG; double TauG;
	vector <double> vec_cpG; double CpG;
	vec_tauG.push_back(0.9);
	vec_cpG.push_back(0.7);
	int DimcpG = vec_cpG.size();
	int DimtG = vec_tauG.size();

	//======================================
	//for hybrid approach
	#ifdef Hybrid_EPSO
		vector <double> vec_tauP; double TauP;
		vector <double> vec_cpP; double CpP;
		vec_tauP.push_back(0.8);
		vec_cpP.push_back(1.0);
		int DimtP = vec_tauP.size();
		int DimcpP = vec_cpP.size();
	#endif

	//======================================
	//Definitions for the cubes
	//Cubos de 3 dimenses (numero iteraes, parametros, seeds)
	int dimCombinationsEPSO = DimcpG*DimtG;
	cube FitsEPSO (bigMaxIterI, 2*dimCombinationsEPSO, SeedDim);
	FitsEPSO.fill(10e+63);
	cube BestSolEPSO (DimIn, 2*dimCombinationsEPSO, SeedDim);
	BestSolEPSO.fill(10e+63);

	//======================================
	//cubes for hybrid
	#ifdef Hybrid_EPSO
		int dimCombinationsHyb = DimcpG*DimcpP*DimtG*DimtP;
		cube FitsHybrid (bigMaxIterI, 2*dimCombinationsHyb, SeedDim);
		FitsHybrid.fill(10e+63);
		cube BestSolHybric (DimIn, 2*dimCombinationsHyb, SeedDim);
		BestSolHybric.fill(10e+63);
	#endif
	
	
	//======================================
	//loop over seeds
	for (unsigned int gs=0; gs<SeedDim; gs++)
	{
		int GlobalSeed=vecGlobalSeed[gs];
		rng.seed(GlobalSeed);
		
		//======================================
		//loop over TauG
		for (unsigned int tg=0; tg<DimtG; tg++)
		{
			//======================================
			//loop over CpG
			for (unsigned int cg=0; cg<DimcpG; cg++)
			{
				TauG=vec_tauG[tg]; //o valor de tau do espao grande  a posio tg do respectivo vector
				CpG=vec_cpG[cg];  //o valor de cp do espao grande  a posio cg do respectivo vector
				
				//########################################### 1ST EPSO #####################################################
				HWswarm primeiroSwarm (
					rng, 
					numParticles, 
					variaveisProblema,  
					typeperiodEntry, 
					windgenEntry, 
					xprefinalEntry, 
					xfinalEntry,
					capmin, 
					capmax, 
					inflow, 
					numres, 
					qmin, 
					qmax, 
					minimize, 
					myFf,  
					CpG, 
					TauG,
					ProblemSelect);

				vec FitsArmadillo(ConjuntoDim); //Tr+Te+Vald
				vec SumArmadillo(ConjuntoDim);
				if(minimize==true)
				{
					for(unsigned int tt=0; tt<(ConjuntoDim); tt++)
					{
						FitsArmadillo(tt)=10e+63;
						SumArmadillo(tt)=10e+63;
					}
				}
				else
				{
					for(unsigned int tt=0; tt<(ConjuntoDim); tt++)
					{
						FitsArmadillo(tt)=-10e+63;
						SumArmadillo(tt)=-10e+63;
					}
				}

				unsigned int number = 0; 
				int cont = 0; //contador do numero de patterns (no fim tem de ter pelo menos TestDim + TrainDim)+Vald
				mat TrainTestVal (ConjuntoDim, DimIn);

				while (number!=bigMaxIterI)
				{
					number++;

					FitsEPSO(number-1, 2*(tg*DimcpG+cg), gs)=number; //cubo de 3 dimenses:: introduz o cabealho da dimensao dos parametros
					FitsEPSO(number-1, 2*(tg*DimcpG+cg)+1, gs)=primeiroSwarm.HWGetBestFitness(); //adiciona o fitness desta iteraao no cubo
					
					//=================================
					//filling the positions within the hybrid cubes in space S (part A)
					//Espao S, antes de treinar rede fica = FitsEPSO
					#ifdef Hybrid_EPSO
						if (number<=maxIterI)
						{
							for (unsigned int dd=0; dd<DimtP*DimcpP; dd++)
							{
								FitsHybrid(number-1, 2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+dd), gs)=number; //cubo dos fitness do hybrid :: regista o numero da iterao
								FitsHybrid(number-1, 2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+dd)+1, gs)=primeiroSwarm.HWGetBestFitness(); //cubo dos fitness do hybrid :: regista o valor do fitness desta iteraao
							}
						}
					#endif

					cout << " i:" << number << " f:" << primeiroSwarm.HWGetBestFitness()<< " Euros"  << endl; 
					
					primeiroSwarm.NextIterationPrimeiro(
						rng,
						number, 
						maxIterI, 
						maxIterS,
						bigMaxIterI,
						variaveisProblema, typeperiodEntry, windgenEntry, xprefinalEntry, xfinalEntry, 
						capmin, 
						capmax, 
						inflow, 
						numres, 
						qmin, 
						qmax, 
						trainDim, 
						TBegin, 
						TEnd, 
						testDim, 
						CpG,
						TauG,
						ProblemSelect,
						Vel_init);
					
					//=================================
					//the selection of the particles is made only for the hybrid case
					//this selections is made following two criteria: the i) best and ii) different particles 
					#ifdef Hybrid_EPSO
						int worst;
						double Fits_actual; int auxFit; 
						double Sum_actual;

						#ifdef TT_SetsSelection_EPSO
						if ((number>=TBegin) && (number <TEnd)) //TEnd = maxIterI; Iteraoes de seleccao dos pattern p tr, te e vald
						{
							list<HWparticle>::iterator pParticle;
							pParticle=primeiroSwarm.EnxameParteA.begin();
							while(pParticle!=primeiroSwarm.EnxameParteA.end())
							{
								double numtemp=0.0;
								uvec indicesFit = sort_index(FitsArmadillo);
								Fits_actual = (pParticle->HWGetFitness());
								Sum_actual=0.0; //coloca o somatorio a zeros
								for (unsigned int gg=0; gg< DimIn; gg++) //calcula o somatorio da particula actual
								{
									Sum_actual+=abs(pParticle->HWGetPos()[gg]);
								}

								auxFit=0;
								for (unsigned int tt=0; tt<(ConjuntoDim); tt++) //percorre todo o conjunto de treino + teste
								{
									/*if(Fits_actual == FitsArmadillo(tt))
									{
										auxFit+=1;//soma 1 qd as particulas tm mm Fitness --> particulas sao =s
									}*/

									if(Sum_actual == SumArmadillo(tt))
									{
										auxFit+=1;//soma 1 qd as particulas tm mm somatrio --> particulas sao =s
									}
								}
								if(auxFit<1) //garante que as particulas sao sempre diferentes,ie, particula actual diferente de todas as outras
								{
									if(minimize==true) //Caso minimizar
									{
										worst = indicesFit(ConjuntoDim-1); //a posio do pior elemento do vetor fit  o maximo
										if ( Fits_actual < FitsArmadillo(worst) )//Fits_actual melhor que o pior --> substituimos
										{
											cont+=1;
											for(unsigned int in=0; in < DimIn; in++)
											{
												numtemp=pParticle->HWGetPos()[in];
												TrainTestVal(worst,in) = numtemp;
											}
											FitsArmadillo(worst) = Fits_actual;
											SumArmadillo(worst) = Sum_actual;
										}
									}
									else //caso maximizar
									{
										worst = indicesFit(0); //a posio do pior elemento do vetor fit  o minimo
										if ( Fits_actual > FitsArmadillo(worst) )
										{
											cont+=1;
											for(unsigned int in=0; in< DimIn; in++)
											{
												numtemp=pParticle->HWGetPos()[in];
												TrainTestVal(worst,in) = numtemp;
											}
											FitsArmadillo(worst) = Fits_actual;
											SumArmadillo(worst) = Sum_actual;
										}
									}
								}
								pParticle++;
							}
							cout << "cont:    " << cont << endl;
						}
						#endif
					#endif
					
				}

				//=============================
				#ifdef Hybrid_EPSO
					//stop the program in case the number of particles are not sufficient
					if (cont < (ConjuntoDim)) 
					{
						cerr << "Error: No Ha particulas suficientes para fazer o Test e o Train SET!!!! AHAHAH :P" << endl; 
						exit(1);
					}
				
					//for the case the random initialization of particles was selected
					//obtemos a matriz conjunta de treino e de test a partir de numeros aleatorios###
					#ifdef TT_SetsSelection_EPSO
					//nada
					#else
						//preenche a matrix TestTrain com numeros aleatores que respeitam o dominio
						//mat TestTrain ((ConjuntoDim), DimIn);
						boost::random::uniform_real_distribution<> aleatTestTrain(minPos[0],maxPos[0]); //neste caso os limites os limites nao sao fixos, nao pode ser assim.

						for (unsigned int i =0; i< (ConjuntoDim); i++)
						{
							for (unsigned int j=0; j< DimIn; j++)
							{
								TrainTestVal (i,j)= aleatTestTrain(rng);
							}
						}
					#endif
				#endif
					
				for (unsigned int in=0; in<DimIn; in++)
				{
					BestSolEPSO(in,2*(tg*DimcpG+cg),gs) = in;
					BestSolEPSO(in,2*(tg*DimcpG+cg)+1,gs) = primeiroSwarm.HWGetBestPos()[in];
				}

				#ifdef Hybrid_EPSO
					// Dados para treino da rede
					vec FitsArmadilloTrain (trainDim); //vector que armazena os valores de fitness relativos ao conjunto de treino
					vec FitsArmadilloTest (testDim); //vector que armazena os valores de fitness relativos ao conjunto de teste
					vec FitsArmadilloValidation (validationDim); 

					//separao train e test para matrizes separadas::: metodo pode ser melhorado
					//saving in a matrix the best testDim particles
					mat notNormTrainSet(trainDim,DimIn); // train set matrix
					mat notNormTestSet(testDim,DimIn);
					mat notNormValidationSet(validationDim,DimIn);////////ATTENTION!!!!!!!

					vec dimensions(3);
					dimensions << trainDim << testDim << validationDim;
					/*dimensions(0)=testDim;
					dimensions(1)=trainDim;
					dimensions(2) = validationDim;*/

					double numpartes = ConjuntoDim / min(dimensions);
				
					int numPartesTr = boost::math::iround((static_cast<double>(trainDim)/static_cast<double>(ConjuntoDim))*numpartes);
					unsigned int numPartesTe = boost::math::iround((static_cast<double>(testDim)/static_cast<double>(ConjuntoDim))*numpartes);
					unsigned int numPartesVal = boost::math::iround((static_cast<double>(validationDim)/static_cast<double>(ConjuntoDim))*numpartes);
					
					unsigned int position = 0;
					unsigned int countTr = 0, countTe = 0, countVal = 0;
				
					while (position < ConjuntoDim)//escolha alternada dos patterns para os conj tr, te e vald
					{//ok
						for (unsigned int tr=0; tr<numPartesTr; tr++)
						{
							if (position == ConjuntoDim || countTr == trainDim)
							{
								break;
							}
							notNormTrainSet(countTr,span()) = TrainTestVal(position,span());
							FitsArmadilloTrain(countTr) = FitsArmadillo(position);
							countTr++;
							position++;
						}

						for (unsigned int te=0; te<numPartesTe; te++)
						{
							if (position == ConjuntoDim || countTe==testDim) // ******ATENTION!!!!!
							{
								break;
							}
							notNormTestSet(countTe,span()) = TrainTestVal(position,span());
							FitsArmadilloTest(countTe) = FitsArmadillo(position);
							countTe++;
							position++;
						}

						for (unsigned int val=0; val<numPartesVal; val++)
						{
							if (position == ConjuntoDim || countVal==validationDim) // ******ATENTION!!!!!
							{
								break;
							}
							notNormValidationSet(countVal,span()) = TrainTestVal(position,span());
							FitsArmadilloValidation(countVal) = FitsArmadillo(position);
							countVal++;
							position++;
						}
					}

				
					//########################################### NETWORK #####################################################
					// ##### Data normalization #####
				
					//MINMAX: Parameter1 = MaxTrain; Z-SCORE: Parameter1 = MeanTrain;
					rowvec Parameter1 =  Normalization::Parameter1_function (notNormTrainSet,  NormMethod, DimIn); //extract the vector with maximum values of TRAIN set
					//MINMAX: Parameter2 = MinTrain; Z-SCORE: Parameter2 = SDev;
					rowvec Parameter2 = Normalization::Parameter2_function (notNormTrainSet,  NormMethod, DimIn); //extract the vector with maximum values of TEST set
				
					const mat TrainSet = Normalization::Normalization_function ( Parameter1,  Parameter2,  notNormTrainSet,  trainDim,  DimIn,  NormMethod); // normalize TRAIN set (using max and min of TRAIN set)
					const mat TestSet = Normalization::Normalization_function (Parameter1,  Parameter2,  notNormTestSet, testDim,  DimIn,  NormMethod); //read & normalize TEST set (using max and min of TRAIN set)
					const mat ValidationSet = Normalization::Normalization_function (Parameter1, Parameter2, notNormValidationSet, validationDim, DimIn, NormMethod); //read & normalize VALIDATION set (using max and min of TRAIN set)
	
					const mat TargetTrSet = TrainSet;
					const mat TargetTsSet = TestSet;
					const mat TargetVlSet = ValidationSet;
				
					mat  W1; rowvec B1; mat W2; rowvec B2;// double Best_MSE_test_Main;
					cout << " Instantiated ! " << endl;

					// ##### Network Training #####
					switch (Train_Global)
					{
						case PROP: 
						{
							NetworkPROP_function(
								W1, B1,W2, B2, 
								Init_W1_Method, 
								Init_W2_Method,
								Init_B1_Method,
								Init_B2_Method,
								Label_data, 
								NeuronActFunct1,
								NeuronActFunct2,
								OjaDecision,
								GlobalSeed,
								rng,
								DimIn, 
								DimHidden, 
								DimOut, 
								trainDim, 
								testDim, 
								validationDim, 
								learnCoeff1,
								TrainSet, 
								TestSet, 
								ValidationSet, 
								TargetTrSet,
								TargetTsSet,
								TargetVlSet,
								numEpochs1, 
								lr_dec, 
								lr_inc, 
								max_perf_inc, 
								goal,
								momRate, 
								actFcnFlatness, 
								SatLimit, 
								ThresholdSat);
							break;
						}
		
						case ITL: 
						{
							NetworkFunction (
								W1, B1,W2, B2, 
								W1_Method, 
								W2_Method,
								Init_W1_Method,
								Init_W2_Method,
								Init_B1_Method,
								Init_B2_Method,
								Label_data,
								sigmaMethod,
								NeuronActFunct1,
								NeuronActFunct2,
								OjaDecision,
								GlobalSeed,
								rng,
								DimIn, 
								DimHidden, 
								DimOut, 
								trainDim, 
								testDim, 
								validationDim,
								learnCoeff1,
								learnCoeff2, 
								TrainSet,
								TestSet,
								ValidationSet,
								TargetTrSet,
								TargetTsSet,
								TargetVlSet,
								numEpochs1,
								numEpochs2,
								lr_dec, //for the learning adaptive decreasing rate (see Hagan book)
								lr_inc, //for the learning adaptide increasing rate (see Hagan book)
								max_perf_inc, //maximum performance increase racio (=CostFunction/Prev_CostFunction) a partir do qual nao se actualizam pesos
								goal, //como no matlab
								momRate, //momentum coef
								actFcnFlatness, // For Activation Function
								SatLimit, //saturation limit 
								ThresholdSat, //saturation threshold
								AuxSigma,
								AuxSigma2,//for sigma_corr
								L,//for sigma_corr
								lrCoeff_sigma//for sigma_corr
								);

							break;
						}

						case PCA_train: 
						{
							General_functions::PCA_function(W1, TrainSet, DimIn, DimHidden);
							W2 = trans(W1);
							B1.resize(DimHidden); B1.zeros();
							B2.resize(DimOut); B2.zeros();
							break;
						}
			
						default: cout << "ERROR!!! ";
					}
						
					cout << endl << " After training ... " << endl;
					cout << endl << " Done!" << endl;
													
					// ######################## 2nd EPSO Small Space ########################

					for (unsigned int tp=0; tp<DimtP; tp+=1) 
					{//ok
						for (unsigned int cp=0; cp<DimcpP; cp+=1)
						{//ok
							TauP=vec_tauP[tp];
							CpP=vec_cpP[cp];

							//criar o swarm para o espao reduzido
							swarm segundoSwarm (
								variaveisProblema,  
								typeperiodEntry, 
								windgenEntry, 
								xprefinalEntry, 
								xfinalEntry,
								capmin, 
								capmax, 
								inflow, 
								numres, 
								qmin, 
								qmax, 
								primeiroSwarm.lastPrimeiroSwarm, 
								primeiroSwarm.beforeLastPrimeiroSwarm,
								numParticles, 
								W1, W2, B1, B2, Parameter1, Parameter2,
								actFcnFlatness,
								NeuronActFunct1,
								NeuronActFunct2,
								NormMethod,
								minimize, 
								myFfs, 
								rng, 
								CpP,
								TauP,
								ProblemSelect, Vel_init);
	
							// Se as 3 ultimas iteracoes retornarem igual valor de Fits ento
							// considera-se que a optimizaao est estagnada e pra o ciclo
							//vec LastThreeFits (3); 
							//LastThreeFits(0)=8e+63; LastThreeFits(1)=4e+63;  LastThreeFits(2)=3e+63;
							//unsigned int comparative=0;
							unsigned int numberSegundo=0;

							while (numberSegundo!=maxIterS /* && comparative == 0*/)
							{//ok
								numberSegundo++;
								cout << " i:" << numberSegundo << " f:" << segundoSwarm.GetBestFitness()<< " Euros"  << endl; 

								FitsHybrid(maxIterI+numberSegundo-1, 2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+tp*DimcpP+cp), gs) = numberSegundo;
								FitsHybrid(maxIterI+numberSegundo-1, 2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+tp*DimcpP+cp)+1, gs)=segundoSwarm.GetBestFitness();
							
								segundoSwarm.NextIterationSegundo(
									variaveisProblema,  
									typeperiodEntry, 
									windgenEntry, 
									xprefinalEntry, 
									xfinalEntry,
									capmin, 
									capmax, 
									inflow, 
									numres, 
									qmin, 
									qmax, 
									rng, 
									W1, W2, B1, B2, Parameter1, Parameter2,
									actFcnFlatness,
									NeuronActFunct2,
									NormMethod,
									maxIterS, 
									numberSegundo,
									DimHidden,
									CpP,
									TauP,
									ProblemSelect);
								/*
								LastThreeFits(2)=LastThreeFits(1);
								LastThreeFits(1)=LastThreeFits(0);
								LastThreeFits(0)=segundoSwarm.GetBestFitness();

								if (LastThreeFits(1)==LastThreeFits(0))
								{//ok
									if (LastThreeFits(1)==LastThreeFits(2))
									{
										comparative = 1;
										//number++;
										// usa swarm da iteraao number+1
										segundoSwarm.NextIterationSegundo(
											variaveisProblema,  
											typeperiodEntry, 
											windgenEntry, 
											xprefinalEntry, 
											xfinalEntry,
											capmin, 
											capmax, 
											inflow, 
											numres, 
											qmin, 
											qmax, 
											rng, 
											W1, W2, B1, B2, Parameter1, Parameter2,
											actFcnFlatness,
											NeuronActFunct2,
											NormMethod,
											numberSegundo+1, //atingiu o criterio de convergencia, e o nr max de iteraoes foi alterado para o nr actual +1
											numberSegundo+1,
											DimHidden,
											CpP,
											TauP,
											ProblemSelect);
									}
								}*/
							}
						
							//At this point, the Last Swarm from the reduced space is available!Now, he must be expand...
							//Idea:pass this lastMiniSwarm to swarmHW.cpp and do a procedure simillar to the creation of the small swarm
	
							//########################################### 3RD EPSO Big Space #####################################################
							HWswarm terceiroSwarm(
								rng, 
								segundoSwarm.lastSegundoSwarm, 
								segundoSwarm.beforeLastSegundoSwarm,
								numParticles, 
								variaveisProblema, typeperiodEntry, windgenEntry, xprefinalEntry, xfinalEntry, 
								capmin, capmax, inflow, numres, qmin, qmax, 
								W2,B2, Parameter1, Parameter2, actFcnFlatness, NeuronActFunct2, NormMethod,
								minimize, 
								myFf,  
								CpG, 
								TauG,
								ProblemSelect,
								Vel_init);


							maxIterF=bigMaxIterI-maxIterI-numberSegundo;//TotalIt
							int auxNumber=maxIterI+numberSegundo;
							unsigned int numberTerceiro=0; 
						
							while (numberTerceiro!=maxIterF)
							{//ok
								numberTerceiro++;
								cout << " i:" << numberTerceiro << " f:" << terceiroSwarm.HWGetBestFitness()<< " Euros"  << endl;

								FitsHybrid(auxNumber+numberTerceiro-1, 2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+tp*DimcpP+cp), gs)=numberTerceiro;
								FitsHybrid(auxNumber+numberTerceiro-1, 2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+tp*DimcpP+cp)+1, gs)=terceiroSwarm.HWGetBestFitness();
							
								//int coisa=0;
								terceiroSwarm.NextIterationTerceiro(
									rng, 
									numberTerceiro, 
									maxIterF, 
									variaveisProblema, typeperiodEntry, windgenEntry, xprefinalEntry, xfinalEntry, 
									capmin, 
									capmax, 
									inflow, 
									numres, 
									qmin, 
									qmax, 
									CpG,
									TauG,
									ProblemSelect,
									Vel_init);
							}

							for (unsigned int in=0; in<DimIn; in++)
							{
								BestSolHybric(in,2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+tp*DimcpP+cp),gs) = in;
								BestSolHybric(in,2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+tp*DimcpP+cp)+1,gs) = terceiroSwarm.HWGetBestPos()[in];
							}

							cout<<"fim"<<endl;
												
							// print the number of iterations		
							cout << "iterations: " << numberTerceiro << " fitness evals: " << terceiroSwarm.HWGetNumEvalsParteC() << " fitbest: " << terceiroSwarm.HWGetBestFitness() << " Euros"  << endl << endl;
							cout << "Parte A num Evals: " << primeiroSwarm.HWGetNumEvalsParteA() << endl;
							cout << "Parte B num Evals: " << segundoSwarm.GetNumEvalsParteB() << endl;
							cout << "Parte C num Evals: " << terceiroSwarm.HWGetNumEvalsParteC() << endl;

							// display the final best solution on the screen
							cout << endl << endl << " ==== best solution found ==== " << endl;
						}
					}
				#endif
			}
		}
	}
	

	//================================
	// string names for outputs
	vector <string> namesEPSO;
	vector <string> namesBestSolEPSO;

	#ifdef Hybrid_EPSO
		vector <string> namesHyb;
		vector <string> namesBestSolHyb;
	#endif

	for (unsigned int tg=0; tg<DimtG; tg++)
	{
		for (unsigned int cg=0; cg<DimcpG; cg++)
		{
			ostringstream tauG_string;
			ostringstream cpG_string;

			TauG = vec_tauG[tg];
			CpG = vec_cpG[cg];
			tauG_string << TauG;
			cpG_string << CpG;

			string str1f = tauG_string.str();//str1f = valor tauG
			string str2f = cpG_string.str();//str2f = valor cpG

			#ifdef Hybrid_EPSO
				for (unsigned int tp=0; tp<DimtP; tp++)
				{
					for (unsigned int cp=0; cp<DimcpP; cp++)
					{
						ostringstream tauP_string;
						ostringstream cpP_string;
										
						TauP = vec_tauP[tp];
						CpP = vec_cpP[cp];
										
						tauP_string << TauP;
						cpP_string << CpP;
										
						string str3f = tauP_string.str();//str3f = valor tauP
						string str4f = cpP_string.str();//str4f = valor cpP

						string matHybrid ("MatHyb_tauG_");
						matHybrid.append(str1f);
						matHybrid.append("_cpG_");
						matHybrid.append(str2f);
						matHybrid.append("_tP_");
						matHybrid.append(str3f);
						matHybrid.append("_cpP_");
						matHybrid.append(str4f);
						matHybrid.append(".csv");

						namesHyb.push_back(matHybrid);

						string matBestSolHyb ("MatBestSolHyb_tauG_");	
						matBestSolHyb.append(str1f);
						matBestSolHyb.append("_cpG_");
						matBestSolHyb.append(str2f);
						matBestSolHyb.append("_tP_");
						matBestSolHyb.append(str3f);
						matBestSolHyb.append("_cpP_");
						matBestSolHyb.append(str4f);
						matBestSolHyb.append(".csv");

						namesBestSolHyb.push_back(matBestSolHyb);
					}
				}
			#endif

			string matEPSO ("MatEPSO_tauG_");
			matEPSO.append(str1f);
			matEPSO.append("_cpG_");
			matEPSO.append(str2f);
			matEPSO.append(".csv");

			namesEPSO.push_back(matEPSO);

			string matBestSolEPSO ("MatBestSolEPSO_tauG_");
			matBestSolEPSO.append(str1f);
			matBestSolEPSO.append("_cpG_");
			matBestSolEPSO.append(str2f);
			matBestSolEPSO.append(".csv");

			namesBestSolEPSO.push_back(matBestSolEPSO);
		}
	}
	// ##### END string names for outputs #####
	

	// ##### Output Files #####
	int contH=0;//contador hyb
	int contE=0;//contador EPSO
	ofstream FitsEPSoOutputs;
	ofstream BestPosEPSoOutputs;

	for (int tg=0; tg<DimtG; tg++)
	{
		for (int cg=0; cg<DimcpG; cg++)
		{
			TauG = vec_tauG[tg];
			CpG = vec_cpG[cg]; 
			
			//=====open
			FitsEPSoOutputs.open(namesEPSO[contE], ios::app);
			BestPosEPSoOutputs.open(namesBestSolEPSO[contE], ios::app);
			for (unsigned int gs=0; gs<SeedDim; gs++)
			{
				int GlobalSeed=vecGlobalSeed[gs];
				FitsEPSoOutputs << "it;"<<GlobalSeed<< ";";
				BestPosEPSoOutputs <<"it;"<<GlobalSeed<< ";";
			}
			FitsEPSoOutputs << "average;" << endl; // cria vector aleatorio dos Fits EPSO
			BestPosEPSoOutputs << "average;" << endl;
			
			FitsEPSoOutputs.close();
			BestPosEPSoOutputs.close();
			//=====closed

			//=====open
			FitsEPSoOutputs.open(namesEPSO[contE], ios::app);
			for (unsigned int it=0; it<bigMaxIterI; it++)
			{
				vec auxmeanEPSO (SeedDim);
				for (unsigned int gs=0; gs<SeedDim; gs++)
				{
					FitsEPSoOutputs << FitsEPSO(it,2*(tg*DimcpG+cg),gs) << ";" << FitsEPSO(it, 2*(tg*DimcpG+cg)+1,gs) << ";";
					auxmeanEPSO(gs) = FitsEPSO(it, 2*(tg*DimcpG+cg)+1,gs); //vector das mdias (auxiliar para ter a coluna das medias a sair nos files finais)
				}
				FitsEPSoOutputs << mean(auxmeanEPSO) << endl;
			}
			FitsEPSoOutputs.close();
			//=====closed

			//=====open
			BestPosEPSoOutputs.open(namesBestSolEPSO[contE], ios::app);
			for (unsigned int in=0; in<DimIn; in++)
			{
				for (unsigned int gs=0; gs<SeedDim; gs++)
				{
					BestPosEPSoOutputs << BestSolEPSO(in,  2*(tg*DimcpG+cg) ,gs) << ";" << BestSolEPSO(in, 2*(tg*DimcpG+cg)+1,gs) << ";";
				}
				BestPosEPSoOutputs << endl;
			}
			BestPosEPSoOutputs.close();
			//=====closed

			#ifdef Hybrid_EPSO
				ofstream FitsHybridOutputs;
				ofstream BestPosHybridOutputs;
				for (unsigned int tp=0; tp<DimtP; tp++)
				{
					for (unsigned int cp=0; cp<DimcpP; cp++)
					{
						TauP=vec_tauP[tp];
						CpP=vec_cpP[cp];
						
						//=====open
						FitsHybridOutputs.open(namesHyb[contH], ios::app);
						BestPosHybridOutputs.open(namesBestSolHyb[contH], ios::app);

						for (unsigned int gs=0; gs<SeedDim; gs++)
						{
							int GlobalSeed=vecGlobalSeed[gs];
							FitsHybridOutputs <<"it;"<<GlobalSeed<< ";";
							BestPosHybridOutputs << "it;"<<GlobalSeed<< ";";
						}
												
						FitsHybridOutputs << "average;" << endl;
						BestPosHybridOutputs << "average;" << endl;

						FitsHybridOutputs.close();
						BestPosHybridOutputs.close();
						//=====closed
						
						//=====open
						FitsHybridOutputs.open(namesHyb[contH], ios::app);

						for (unsigned int it=0; it<bigMaxIterI; it++)
						{
							vec auxmeanHyb (SeedDim);
							for (unsigned int gs=0; gs<SeedDim; gs++)
							{
								FitsHybridOutputs << FitsHybrid(it, 2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+tp*DimcpP+cp),gs) << ";" << FitsHybrid(it,2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+tp*DimcpP+cp)+1,gs) << ";";
								auxmeanHyb(gs) = FitsHybrid(it, 2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+tp*DimcpP+cp)+1,gs); //vector das mdias (auxiliar para ter a coluna das medias a sair nos files finais)
							}
							FitsHybridOutputs << mean(auxmeanHyb) << endl;
						}
						FitsHybridOutputs.close();
						//=====closed

						//=====open
						BestPosHybridOutputs.open(namesBestSolHyb[contH], ios::app);
						for (unsigned int in=0; in<DimIn; in++)
						{
							for (unsigned int gs=0; gs<SeedDim; gs++)
							{
								BestPosHybridOutputs << BestSolHybric(in, 2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+tp*DimcpP+cp),gs) << ";" << BestSolHybric(in,2*(tg*DimcpG*DimtP*DimcpP+cg*DimtP*DimcpP+tp*DimcpP+cp)+1,gs) << ";";
							}
							BestPosHybridOutputs << endl;
						}
						
						BestPosHybridOutputs.close();
						//=====closed

						contH++;
					}
				}
			#endif
			
			contE++;
		}
	}
	// ##### END of Output Files #####
	
	return;
}
